# ruff: noqa: ARG001
from __future__ import annotations

import copy
import logging
from datetime import datetime, timedelta
from pathlib import Path

import bokeh.palettes as pal
import colorcet as cc
import numexpr as ne
import numpy as np
import pandas as pd
from bokeh.models import (
    CustomJSTickFormatter,
    FixedTicker,
    Label,
    Span,
    ZoomInTool,
    ZoomOutTool,
)
from bokeh.plotting import figure
from dbetto import Props
from legendmeta import LegendMetadata

from legenddashboard.geds.string_visulization import create_detector_plot
from legenddashboard.util import sorter

log = logging.getLogger(__name__)


def build_string_array(chan_map):
    dets = []
    strings = []
    positions = []
    for key, entry in chan_map.items():
        if entry.system == "geds":
            string = entry.location.string
            pos = entry.location.position
            dets.append(key)
            strings.append(string)
            positions.append(int(pos))

    return dets, strings, positions


def build_status_map(chan_map, data):
    dets, strings, positions = build_string_array(chan_map)

    string_nos = np.array(sorted(np.unique(strings)))
    pos_nos = np.array(sorted(np.unique(positions)))
    n_strings = len(string_nos)
    max_pos = np.max(positions)

    data_array = np.full((max_pos * 2 + 1, n_strings * 2 + 1), np.nan)
    annot_array = np.empty((max_pos * 2 + 1, n_strings * 2 + 1), dtype="object")

    for i, det in enumerate(dets):
        index = (
            2 * positions[i] - 1,
            2 * (np.where(strings[i] == string_nos)[0] + 1) - 1,
        )
        annot_array[index] = det
        proc_status = None
        use_status = None
        proc_status = data[det]["processable"]
        use_status = data[det]["usability"]
        if proc_status:
            if use_status == "On":
                data_array[index] = 2.0
            else:
                data_array[index] = 1.0
        else:
            data_array[index] = 0.0

    x_axes = np.full(n_strings * 2 + 1, " ", dtype=object)
    for i, s in enumerate(string_nos):
        x_axes[2 * (i + 1) - 1] = f"Str {s}"

    y_axes = np.full(max_pos * 2 + 1, " ", dtype=object)
    for i, n in enumerate(pos_nos):
        y_axes[2 * (i + 1) - 1] = f"Pos {n}"

    return data_array, x_axes, y_axes, annot_array


def plot_status(
    prod_config,
    run,
    run_dict,
    path,
    source,
    xlabels,
    period,
    key=None,
    sort_dets_obj=None,
    cache_data=None,
):
    if sort_dets_obj is not None:
        statuses = sort_dets_obj.statuses.valid_for(run_dict["timestamp"])
        cmap = sort_dets_obj.chmaps.valid_for(run_dict["timestamp"])
    else:
        chmap = LegendMetadata(path=prod_config["paths"]["chan_map"])
        cmap = chmap.channelmap(run_dict["timestamp"])
        status_map = LegendMetadata(
            path=prod_config["paths"]["detector_status"], lazy=True
        ).statuses
        statuses = status_map.on(run_dict["timestamp"])

    dets, strings, positions = build_string_array(cmap)

    color_dict = {"on": 2, "off": 0, "ac": 1}
    display_dict = {i: color_dict[statuses[i]["usability"]] for i in dets}

    # display_dict = {cmap[i]['daq']['rawid'] : 1 if status_map[i]["processable"] == True and status_map[i]["usability"] == 'on' else 0
    #     for i in dets}
    palette = ("red", "orange", "green")
    ctitle = "Detector Status"
    ticker = FixedTicker(ticks=[0.3, 1.0, 1.7], tags=["red", "orange", "green"])
    formatter = CustomJSTickFormatter(
        code="""
        var mapping = {0.3: "off", 1.0: "ac", 1.7: "on"};
        return mapping[tick];
    """
    )
    return create_detector_plot(
        source,
        display_dict,
        xlabels,
        ctitle=ctitle,
        palette=palette,
        ticker=ticker,
        formatter=formatter,
        plot_title=f"{run_dict['experiment']}-{period}-{run} | Cal. | Detector Status",
        boolean_scale=True,
    )


def build_counts_map(chan_map, data):
    dets, strings, positions = build_string_array(chan_map)

    string_nos = np.array(sorted(np.unique(strings)))
    pos_nos = np.array(sorted(np.unique(positions)))
    n_strings = len(string_nos)
    max_pos = np.max(positions)

    data_array = np.full((max_pos * 2 + 1, n_strings * 2 + 1), np.nan)
    annot_array = np.empty((max_pos * 2 + 1, n_strings * 2 + 1), dtype="object")

    for i, det in enumerate(dets):
        index = (
            2 * positions[i] - 1,
            2 * (np.where(strings[i] == string_nos)[0] + 1) - 1,
        )
        annot_array[index] = data[det]
        data_array[index] = data[det]

    x_axes = np.full(n_strings * 2 + 1, " ", dtype=object)
    for i, s in enumerate(string_nos):
        x_axes[2 * (i + 1) - 1] = f"Str {s}"

    y_axes = np.full(max_pos * 2 + 1, " ", dtype=object)
    for i, n in enumerate(pos_nos):
        y_axes[2 * (i + 1) - 1] = f"Pos {n}"

    return data_array, x_axes, y_axes, annot_array


def plot_counts(
    prod_config,
    run,
    run_dict,
    path,
    source,
    xlabels,
    period,
    key=None,
    sort_dets_obj=None,
    cache_data=None,
):  # FEP counts plot
    if sort_dets_obj is not None:
        cmap = sort_dets_obj.chmaps.valid_for(run_dict["timestamp"])
    else:
        chmap = LegendMetadata(path=prod_config["paths"]["metadata"])
        cmap = chmap.channelmap(run_dict["timestamp"])

    if cache_data is not None and run in cache_data["hit"]:
        all_res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )
        all_res = Props.read_from(path)
        cache_data["hit"][run] = all_res

    res = {}
    for det in cmap:
        if cmap[det].system == "geds":
            try:
                fep_counts = all_res[det]["results"]["ecal"]["cuspEmax_ctc_cal"][
                    "total_fep"
                ]
                mass = cmap[det]["production"]["mass_in_g"]
                mass_in_kg = mass * 0.001
                counts_per_kg = fep_counts / mass_in_kg  # calculate counts per kg
                round_counts = round(counts_per_kg, 0)
                res[det] = round_counts
            except KeyError:
                res[det] = 0

    display_dict = res
    ctitle = "FEP Counts per kg"
    palette = pal.plasma(256)  # alternatively use viridis palette = viridis(256)
    return create_detector_plot(
        source,
        display_dict,
        xlabels,
        ctitle=ctitle,
        palette=palette,
        plot_title=f"{run_dict['experiment']}-{period}-{run} | Cal. | FEP Counts per kg",
        colour_max=10000,
        colour_min=1000,
    )


def plot_energy_resolutions(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    at="Qbb",
    download=False,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["hit"]:
        all_res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        all_res = Props.read_from(path)
        cache_data["hit"][run] = all_res

    default = {
        "cuspEmax_ctc_cal": {
            "Qbb_fwhm": np.nan,
            "Qbb_fwhm_err": np.nan,
            "2.6_fwhm": np.nan,
            "2.6_fwhm_err": np.nan,
            "m0": np.nan,
            "m1": np.nan,
        },
        "zacEmax_ctc_cal": {
            "Qbb_fwhm": np.nan,
            "Qbb_fwhm_err": np.nan,
            "2.6_fwhm": np.nan,
            "2.6_fwhm_err": np.nan,
            "m0": np.nan,
            "m1": np.nan,
        },
        "trapEmax_ctc_cal": {
            "Qbb_fwhm": np.nan,
            "Qbb_fwhm_err": np.nan,
            "2.6_fwhm": np.nan,
            "2.6_fwhm_err": np.nan,
            "m0": np.nan,
            "m1": np.nan,
        },
    }
    res = {}
    for stri in strings:
        res[stri] = default
        for channel in strings[stri]:
            detector = channel_map[channel]["name"]
            try:
                det_dict = all_res[detector]["results"]["ecal"]
                res[detector] = {
                    "cuspEmax_ctc_cal": {
                        "Qbb_fwhm": det_dict["cuspEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_in_kev"
                        ],
                        "Qbb_fwhm_err": det_dict["cuspEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_err_in_kev"
                        ],
                        "2.6_fwhm": det_dict["cuspEmax_ctc_cal"]["pk_fits"][2614.511][
                            "fwhm_in_kev"
                        ],
                        "2.6_fwhm_err": det_dict["cuspEmax_ctc_cal"]["pk_fits"][
                            2614.511
                        ]["fwhm_err_in_kev"],
                        "m0": det_dict["cuspEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "a"
                        ],
                        "m1": det_dict["cuspEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "b"
                        ],
                    },
                    "zacEmax_ctc_cal": {
                        "Qbb_fwhm": det_dict["zacEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_in_kev"
                        ],
                        "Qbb_fwhm_err": det_dict["zacEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_err_in_kev"
                        ],
                        "2.6_fwhm": det_dict["zacEmax_ctc_cal"]["pk_fits"][2614.511][
                            "fwhm_in_kev"
                        ],
                        "2.6_fwhm_err": det_dict["zacEmax_ctc_cal"]["pk_fits"][
                            2614.511
                        ]["fwhm_err_in_kev"],
                        "m0": det_dict["zacEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "a"
                        ],
                        "m1": det_dict["zacEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "b"
                        ],
                    },
                    "trapEmax_ctc_cal": {
                        "Qbb_fwhm": det_dict["trapEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_in_kev"
                        ],
                        "Qbb_fwhm_err": det_dict["trapEmax_ctc_cal"]["eres_linear"][
                            "Qbb_fwhm_err_in_kev"
                        ],
                        "2.6_fwhm": det_dict["trapEmax_ctc_cal"]["pk_fits"][2614.511][
                            "fwhm_in_kev"
                        ],
                        "2.6_fwhm_err": det_dict["trapEmax_ctc_cal"]["pk_fits"][
                            2614.511
                        ]["fwhm_err_in_kev"],
                        "m0": det_dict["trapEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "a"
                        ],
                        "m1": det_dict["trapEmax_ctc_cal"]["eres_linear"]["parameters"][
                            "b"
                        ],
                    },
                }
            except KeyError:
                res[detector] = default

    p = figure(
        width=1400,
        height=600,
        y_range=(1, 5),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | {at} Energy Resolution"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(res)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    for filter_type in ["cuspEmax_ctc_cal", "zacEmax_ctc_cal", "trapEmax_ctc_cal"]:
        x_plot, y_plot, y_plot_err = (
            np.arange(1, len(list(res)) + 1, 1),
            [res[det][filter_type][f"{at}_fwhm"] for det in res],
            [res[det][filter_type][f"{at}_fwhm_err"] for det in res],
        )

        err_xs = []
        err_ys = []

        for x, y, yerr in zip(x_plot, y_plot, y_plot_err, strict=False):
            err_xs.append((x, x))
            err_ys.append((np.nan_to_num(y - yerr), np.nan_to_num(y + yerr)))

        df_plot["x_{}".format(filter_type.split("_")[0])] = np.nan_to_num(x_plot)
        df_plot["y_{}".format(filter_type.split("_")[0])] = np.nan_to_num(y_plot)
        df_plot["y_{}_err".format(filter_type.split("_")[0])] = np.nan_to_num(
            y_plot_err
        )
        df_plot["err_xs_{}".format(filter_type.split("_")[0])] = err_xs
        df_plot["err_ys_{}".format(filter_type.split("_")[0])] = err_ys

    if download:
        if at == "Qbb":
            return (
                df_plot,
                f"{run_dict['experiment']}-{period}-{run}_Qbb_energy_resolutions.csv",
            )
        return (
            df_plot,
            f"{run_dict['experiment']}-{period}-{run}_FEP_energy_resolutions.csv",
        )

    for filter_type, filter_name, filter_plot_color in zip(
        ["cuspEmax_ctc_cal", "zacEmax_ctc_cal", "trapEmax_ctc_cal"],
        ["Cusp", "ZAC", "Trap"],
        ["blue", "green", "red"],
        strict=False,
    ):
        if filter_name == "Cusp":
            hover_renderer = p.scatter(
                x="x_{}".format(filter_type.split("_")[0]),
                y="y_{}".format(filter_type.split("_")[0]),
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
                name=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
            )
        else:
            p.scatter(
                x="x_{}".format(filter_type.split("_")[0]),
                y="y_{}".format(filter_type.split("_")[0]),
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
                name=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
            )
        p.multi_line(
            xs="err_xs_{}".format(filter_type.split("_")[0]),
            ys="err_ys_{}".format(filter_type.split("_")[0]),
            source=df_plot,
            color=filter_plot_color,
            legend_label=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
            name=f'{filter_name} Average: {np.nanmean([res[det][filter_type][f"{at}_fwhm"] for det in res]):.2f}keV',
        )

    p.legend.location = "top_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    if at == "Qbb":
        p.yaxis.axis_label = "FWHM at Qbb (keV)"
    else:
        p.yaxis.axis_label = "FWHM of 2.6 MeV peak (keV)"
        p.title.text = (
            f"{run_dict['experiment']}-{period}-{run} | Cal. | FEP Energy Resolution"
        )
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(res)) + 1, 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(res)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=1.1, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        ("FWHM Cusp", "@y_cuspEmax{0.00} +- @y_cuspEmax_err{0.00} keV"),
        ("FWHM ZAC ", "@y_zacEmax{0.00} +- @y_zacEmax_err{0.00} keV"),
        ("FWHM Trap", "@y_trapEmax{0.00} +- @y_trapEmax_err{0.00} keV"),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def plot_energy_resolutions_Qbb(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    download=False,
    cache_data=None,
):
    return plot_energy_resolutions(
        prod_config,
        run,
        run_dict,
        path,
        period,
        key=key,
        at="Qbb",
        sort_dets_obj=sort_dets_obj,
        download=download,
        cache_data=cache_data,
    )


def plot_energy_resolutions_2614(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    download=False,
    cache_data=None,
):
    return plot_energy_resolutions(
        prod_config,
        run,
        run_dict,
        path,
        period,
        key=key,
        sort_dets_obj=sort_dets_obj,
        at="2.6",
        download=download,
        cache_data=cache_data,
    )


def plot_energy_residuals(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    filter_param="cuspEmax_ctc_cal",
    download=False,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["hit"]:
        all_res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        all_res = Props.read_from(path)
        cache_data["hit"][run] = all_res

    peaks = [2614.511, 583.191, 2103.511]
    filters = ["cuspEmax_ctc_cal", "zacEmax_ctc_cal", "trapEmax_ctc_cal"]

    default_peaks = {f"{peak}": np.nan for peak in peaks}
    default_peaks.update({f"{peak}_err": np.nan for peak in peaks})
    default = {filt: default_peaks.copy() for filt in filters}
    res = {}
    for stri in strings:
        res[stri] = copy.deepcopy(default)
        for channel in strings[stri]:
            detector = channel_map[channel]["name"]
            res[detector] = copy.deepcopy(default)
            try:
                det_dict = all_res[detector]["results"]["ecal"]
                for filt in filters:
                    cal_dict = all_res[detector]["pars"]["operations"][filt]
                    for peak in peaks:
                        # try:
                        cal_mu = ne.evaluate(
                            f"{cal_dict['expression']}",
                            local_dict=dict(
                                {
                                    filt.replace("_cal", ""): det_dict[filt]["pk_fits"][
                                        peak
                                    ]["parameters"]["mu"]
                                },
                                **cal_dict["parameters"],
                            ),
                        )
                        cal_err = (
                            ne.evaluate(
                                f"{cal_dict['expression']}",
                                local_dict=dict(
                                    {
                                        filt.replace("_cal", ""): det_dict[filt][
                                            "pk_fits"
                                        ][peak]["uncertainties"]["mu"]
                                        + det_dict[filt]["pk_fits"][peak]["parameters"][
                                            "mu"
                                        ]
                                    },
                                    **cal_dict["parameters"],
                                ),
                            )
                            - cal_mu
                        )
                        res[detector][filt][f"{peak}"] = cal_mu - peak
                        res[detector][filt][f"{peak}_err"] = cal_err
                        # except KeyError:
                        #     pass
            except KeyError:
                pass

    p = figure(
        width=1400,
        height=600,
        tools="pan,box_zoom,ywheel_zoom,hover,reset,save",
        active_scroll="ywheel_zoom",
    )

    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | Energy Residuals"
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(res)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    for filter_type in filters:
        for peak in peaks:
            x_plot, y_plot, y_plot_err = (
                np.arange(1, len(list(res)) + 1, 1),
                [res[det][filter_type][f"{peak}"] for det in res],
                [res[det][filter_type][f"{peak}_err"] for det in res],
            )

            err_xs = []
            err_ys = []

            for x, y, yerr in zip(x_plot, y_plot, y_plot_err, strict=False):
                err_xs.append((x, x))
                err_ys.append((np.nan_to_num(y - yerr), np.nan_to_num(y + yerr)))

            df_plot[f"x_{filter_type.split('_')[0]}_{int(peak)}"] = np.nan_to_num(
                x_plot
            )
            df_plot[f"y_{filter_type.split('_')[0]}_{int(peak)}"] = np.nan_to_num(
                y_plot
            )
            df_plot[f"y_{filter_type.split('_')[0]}_{int(peak)}_err"] = np.nan_to_num(
                y_plot_err
            )
            df_plot[f"err_xs_{filter_type.split('_')[0]}_{int(peak)}"] = err_xs
            df_plot[f"err_ys_{filter_type.split('_')[0]}_{int(peak)}"] = err_ys

    if download:
        return df_plot, f"{run_dict['experiment']}-{period}-{run}_energy_residuals.csv"

    for peak, peak_color in zip(peaks, ["blue", "green", "red"], strict=False):
        if peak == peaks[0]:
            hover_renderer = p.scatter(
                x=f"x_{filter_type.split('_')[0]}_{int(peak)}",
                y=f"y_{filter_type.split('_')[0]}_{int(peak)}",
                source=df_plot,
                color=peak_color,
                size=7,
                line_alpha=0,
                legend_label=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
                name=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
            )
        else:
            p.scatter(
                x=f"x_{filter_type.split('_')[0]}_{int(peak)}",
                y=f"y_{filter_type.split('_')[0]}_{int(peak)}",
                source=df_plot,
                color=peak_color,
                size=7,
                line_alpha=0,
                legend_label=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
                name=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
            )
        p.multi_line(
            xs=f"err_xs_{filter_type.split('_')[0]}_{int(peak)}",
            ys=f"err_ys_{filter_type.split('_')[0]}_{int(peak)}",
            source=df_plot,
            color=peak_color,
            legend_label=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
            name=f'{peak} Average: {np.nanmean([res[det][filter_param][f"{peak}"] for det in res]):.2f}keV',
        )

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "peak residuals (keV)"
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | Energy Residuals"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(res)) + 1, 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(res)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=1.1, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        (
            "2614",
            f"@y_{filter_type.split('_')[0]}_2614{{0.00}} +- @y_{filter_type.split('_')[0]}_2614_err{{0.00}} keV",
        ),
        (
            "SEP",
            f"@y_{filter_type.split('_')[0]}_2103{{0.00}} +- @y_{filter_type.split('_')[0]}_2103_err{{0.00}} keV",
        ),
        (
            "583",
            f"@y_{filter_type.split('_')[0]}_583{{0.00}} +- @y_{filter_type.split('_')[0]}_583_err{{0.00}} keV",
        ),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def plot_no_fitted_energy_peaks(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], sort_dets_obj=sort_dets_obj
    )

    if sort_dets_obj is not None:
        chmap = sort_dets_obj.chmaps.valid_for(run_dict["timestamp"])
    else:
        cfg_file = prod_config["paths"]["chan_map"]
        configs = LegendMetadata(path=cfg_file)
        chmap = configs.channelmaps.on(run_dict["timestamp"])
    channels = [field for field in chmap if chmap[field]["system"] == "geds"]

    off_dets = [
        chmap[field] for field in soft_dict if soft_dict[field]["processable"] is False
    ]

    if cache_data is not None and run in cache_data["hit"]:
        res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        res = Props.read_from(path)
        cache_data["hit"][run] = res

    peaks = [583.191, 727.33, 860.564, 1592.511, 1620.5, 2103.511, 2614.511]
    grid = np.ones((len(peaks), len(channels)))
    for i, channel in enumerate(channels):
        idxs = np.zeros(len(peaks), dtype=bool)
        try:
            fitted_peaks = res[channel]["results"]["ecal"]["cuspEmax_ctc_cal"][
                "pk_fits"
            ]
            fitted_peaks = [pk for pk in fitted_peaks if fitted_peaks["pk"]["validity"]]
            if not isinstance(fitted_peaks, list):
                fitted_peaks = [fitted_peaks]
            for j, peak in enumerate(peaks):
                if peak in fitted_peaks:
                    idxs[j] = 1
            if len(idxs) > 0:
                grid[idxs, i] = 2

        except KeyError:
            if channel in off_dets:
                grid[:, i] = 0

    p = figure(
        width=1400,
        height=300,
        y_range=(0, 7),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | Energy fits"
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [f"{chmap[channel]['name']}" for channel in channels]
    p.image(
        image=[grid],
        x=0,
        y=0,
        dw=len(channels),
        dh=len(peaks),
        palette=["red", "orange", "green"],
        alpha=0.7,
    )

    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Peaks"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = FixedTicker(
        ticks=np.arange(0, len(channels), 1) + 0.5,
        minor_ticks=np.arange(0, len(channels), 1),
    )
    # p.xaxis.major_label_overrides = {i: label_res[i-1] for i in range(1, len(label_res)+1, 1)}
    p.xaxis.major_label_overrides = {
        i + 0.5: label_res[i] for i in range(0, len(label_res), 1)
    }
    p.xaxis.major_label_text_font_style = "bold"
    p.xgrid.grid_line_color = None
    p.xgrid.minor_grid_line_color = "black"
    p.xgrid.minor_grid_line_alpha = 0.1
    p.xaxis.minor_tick_line_color = None

    p.yaxis.ticker = FixedTicker(
        ticks=np.arange(0, len(peaks), 1) + 0.5, minor_ticks=np.arange(0, len(peaks), 1)
    )
    p.yaxis.major_label_overrides = {
        i + 0.5: f"{peaks[i]}" for i in range(0, len(peaks), 1)
    }
    p.ygrid.grid_line_color = None
    p.ygrid.minor_grid_line_color = "black"
    p.ygrid.minor_grid_line_alpha = 0.1
    p.yaxis.minor_tick_line_color = None

    return p


def plot_aoe_status(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], sort_dets_obj=sort_dets_obj
    )

    if sort_dets_obj is not None:
        chmap = sort_dets_obj.chmaps.valid_for(run_dict["timestamp"])
    else:
        cfg_file = prod_config["paths"]["chan_map"]
        configs = LegendMetadata(path=cfg_file)
        chmap = configs.channelmaps.on(run_dict["timestamp"])
    channels = [field for field in chmap if chmap[field]["system"] == "geds"]

    off_dets = [
        field for field in soft_dict if soft_dict[field]["processable"] is False
    ]

    if cache_data is not None and run in cache_data["hit"]:
        res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        res = Props.read_from(path)
        cache_data["hit"][run] = res

    checks = ["Time_corr", "Energy_corr", "Cut_det", "Low_side_sfs", "2_side_sfs"]

    grid = np.ones((len(checks), len(channels)))
    for i, channel in enumerate(channels):
        idxs = np.ones(len(checks), dtype=bool)
        try:
            aoe_dict = res[channel]["results"]["aoe"]
            if not np.isnan(aoe_dict["1000-1300keV"]["0"]["mean"]):
                idxs[0] = 1
            if not (
                np.isnan(
                    np.array(
                        [
                            value
                            for key, value in aoe_dict["correction_fit_results"][
                                "mean_fits"
                            ]["pars"].items()
                        ]
                    )
                ).all()
            ):
                idxs[1] = 1
            if not np.isnan(aoe_dict["low_cut"]):
                idxs[2] = 1
            if isinstance(aoe_dict["low_side_sfs"], float):
                pass
            else:
                sfs = [
                    float(dic["sf"]) for peak, dic in aoe_dict["low_side_sfs"].items()
                ]
                if not np.isnan(np.array(sfs)).all():
                    idxs[3] = 1
            if isinstance(aoe_dict["2_side_sfs"], float):
                pass
            else:
                sfs = [float(dic["sf"]) for peak, dic in aoe_dict["2_side_sfs"].items()]
                if not np.isnan(np.array(sfs)).all():
                    idxs[4] = 1

            if len(idxs) > 0:
                grid[idxs, i] = 2

        except KeyError:
            if channel in off_dets:
                grid[:, i] = 0

    p = figure(
        width=1400,
        height=300,
        y_range=(0, 5),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | A/E status"
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [f"{chmap[channel]['name']}" for channel in channels]
    p.image(
        image=[grid],
        x=0,
        y=0,
        dw=len(channels),
        dh=len(checks),
        palette=["red", "orange", "green"],
        alpha=0.7,
    )

    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Peaks"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = FixedTicker(
        ticks=np.arange(0, len(channels), 1) + 0.5,
        minor_ticks=np.arange(0, len(channels), 1),
    )
    # p.xaxis.major_label_overrides = {i: label_res[i-1] for i in range(1, len(label_res)+1, 1)}
    p.xaxis.major_label_overrides = {
        i + 0.5: label_res[i] for i in range(0, len(label_res), 1)
    }
    p.xaxis.major_label_text_font_style = "bold"
    p.xgrid.grid_line_color = None
    p.xgrid.minor_grid_line_color = "black"
    p.xgrid.minor_grid_line_alpha = 0.1
    p.xaxis.minor_tick_line_color = None

    p.yaxis.ticker = FixedTicker(
        ticks=np.arange(0, len(checks), 1) + 0.5,
        minor_ticks=np.arange(0, len(checks), 1),
    )
    p.yaxis.major_label_overrides = {
        i + 0.5: f"{checks[i]}" for i in range(0, len(checks), 1)
    }
    p.ygrid.grid_line_color = None
    p.ygrid.minor_grid_line_color = "black"
    p.ygrid.minor_grid_line_alpha = 0.1
    p.yaxis.minor_tick_line_color = None

    return p


def plot_no_fitted_aoe_slices(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["hit"]:
        res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        res = Props.read_from(path)
        cache_data["hit"][run] = res

    nfits = {}
    for stri in strings:
        nfits[stri] = np.nan
        for channel in strings[stri]:
            detector = channel_map[channel]["name"]
            try:
                nfits[detector] = res[channel]["results"]["aoe"][
                    "correction_fit_results"
                ]["n_of_valid_fits"]
            except KeyError:
                nfits[detector] = np.nan

    p = figure(
        width=1400,
        height=600,
        y_range=(1, 70),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | A/E fits"
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(nfits)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    df_plot["x_nfits"] = np.arange(1, len(list(nfits)) + 1, 1)
    df_plot["nfits"] = np.nan_to_num(list(nfits.values()))

    filter_types = ["nfits"]
    filter_names = ["Valid. A/E fits"]
    filter_plot_colors = ["blue"]

    for filter_type, filter_name, filter_plot_color in zip(
        filter_types, filter_names, filter_plot_colors, strict=False
    ):
        if filter_name == "Valid. A/E fits":
            hover_renderer = p.scatter(
                x=f"x_{filter_type}",
                y=filter_type,
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=filter_name,
                name=filter_name,
            )
        else:
            p.scatter(
                x=f"x_{filter_type}",
                y=filter_type,
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=filter_name,
                name=filter_name,
            )

    p.legend.location = "top_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "# of A/E Fits"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(nfits)) + 1, 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(nfits)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=1.5, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [("Detector", "@label_res"), ("Valid. A/E fits", "@nfits")]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def get_aoe_results(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    download=False,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["hit"]:
        all_res = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_hit.yaml'
        )

        all_res = Props.read_from(path)
        cache_data["hit"][run] = all_res

    default = {
        "A/E_Energy_param": "cuspEmax",
        "Cal_energy_param": "cuspEmax_ctc",
        "dt_param": "dt_eff",
        "rt_correction": False,
        "mean_pars": [np.nan, np.nan],
        "sigma_pars": [np.nan, np.nan],
        "low_cut": np.nan,
        "high_cut": np.nan,
        "low_side_sfs": {
            1592.5: {"sf": np.nan, "sf_err": np.nan},
            1620.5: {"sf": np.nan, "sf_err": np.nan},
            2039.0: {"sf": np.nan, "sf_err": np.nan},
            2103.53: {"sf": np.nan, "sf_err": np.nan},
            2614.5: {"sf": np.nan, "sf_err": np.nan},
        },
        "2_side_sfs": {
            1592.5: {"sf": np.nan, "sf_err": np.nan},
            1620.5: {"sf": np.nan, "sf_err": np.nan},
            2039.0: {"sf": np.nan, "sf_err": np.nan},
            2103.53: {"sf": np.nan, "sf_err": np.nan},
            2614.5: {"sf": np.nan, "sf_err": np.nan},
        },
    }

    aoe_res = {}
    for stri in strings:
        aoe_res[stri] = default
        for channel in strings[stri]:
            detector = channel_map[channel]["name"]

            try:
                aoe_res[detector] = all_res[detector]["results"]["aoe"]
                if len(aoe_res[detector]) == 0:
                    raise KeyError
            except KeyError:
                aoe_res[detector] = default.copy()

    # log.debug(aoe_res["V02160A"])
    # log.debug(aoe_res["P00698B"])

    p = figure(
        width=1400,
        height=600,
        y_range=(-5, 100),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | A/E Survival Fractions"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(aoe_res)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    peak_types = [1592.5, 1620.5, 2039.0, 2103.53, 2614.5]
    peak_names = ["Tl DEP", "Bi FEP", "CC @ Qbb", "Tl SEP", "Tl FEP"]
    peak_colors = ["blue", "orange", "green", "red", "purple"]

    for peak_type in peak_types:
        for det in aoe_res:
            log.debug(det)
            log.debug(aoe_res[det]["low_side_sfs"][peak_type]["sf"])
        # try:
        x_plot, y_plot, y_plot_err = (
            np.arange(1, len(list(aoe_res)) + 1, 1),
            [float(aoe_res[det]["low_side_sfs"][peak_type]["sf"]) for det in aoe_res],
            [
                float(aoe_res[det]["low_side_sfs"][peak_type]["sf_err"])
                for det in aoe_res
            ],
        )

        err_xs = []
        err_ys = []

        for x, y, yerr in zip(x_plot, y_plot, y_plot_err, strict=False):
            err_xs.append((x, x))
            err_ys.append((np.nan_to_num(y - yerr), np.nan_to_num(y + yerr)))

        df_plot["x_{}".format(str(peak_type).split(".")[0])] = np.nan_to_num(x_plot)
        df_plot["y_{}".format(str(peak_type).split(".")[0])] = np.nan_to_num(y_plot)
        df_plot["y_{}_err".format(str(peak_type).split(".")[0])] = np.nan_to_num(
            y_plot_err
        )
        df_plot["err_xs_{}".format(str(peak_type).split(".")[0])] = err_xs
        df_plot["err_ys_{}".format(str(peak_type).split(".")[0])] = err_ys

    if download:
        return (
            df_plot,
            f"{run_dict['experiment']}-{period}-{run}_AoE_SurvivalFractions.csv",
        )

    for peak_type, peak_name, peak_color in zip(
        peak_types, peak_names, peak_colors, strict=False
    ):
        if peak_type == 1592.5:
            hover_renderer = p.scatter(
                x="x_{}".format(str(peak_type).split(".")[0]),
                y="y_{}".format(str(peak_type).split(".")[0]),
                source=df_plot,
                color=peak_color,
                size=7,
                line_alpha=0,
                legend_label=peak_name,
                name=peak_name,
            )
        else:
            p.scatter(
                x="x_{}".format(str(peak_type).split(".")[0]),
                y="y_{}".format(str(peak_type).split(".")[0]),
                source=df_plot,
                color=peak_color,
                size=7,
                line_alpha=0,
                legend_label=peak_name,
                name=peak_name,
            )
        p.multi_line(
            xs="err_xs_{}".format(str(peak_type).split(".")[0]),
            ys="err_ys_{}".format(str(peak_type).split(".")[0]),
            source=df_plot,
            color=peak_color,
            legend_label=peak_name,
            name=peak_name,
        )

    p.legend.location = "top_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Survival fraction (%)"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(aoe_res)), 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(aoe_res)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=-5, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        ("SF CC @Qbb", "@y_2039{0.0} +- @y_2039_err{0.0} %"),
        ("SF Tl DEP", "@y_1592{0.0} +- @y_1592_err{0.0} %"),
        ("SF Bi FEP", "@y_1620{0.0} +- @y_1620_err{0.0} %"),
        ("SF Tl SEP", "@y_2103{0.0} +- @y_2103_err{0.0} %"),
        ("SF Tl FEP", "@y_2614{0.0} +- @y_2614_err{0.0} %"),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def plot_pz_consts(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    download=False,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["dsp"]:
        cal_dict = cache_data["hit"][run]
    else:
        file_path = Path(prod_config["paths"]["par_dsp"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_dsp.yaml'
        )

        cal_dict = Props.read_from(path)
        cache_data["dsp"][run] = cal_dict

    taus = {}

    for stri in strings:
        taus[stri] = np.nan
        for channel in strings[stri]:
            det = channel_map[channel]["name"]
            try:
                taus[det] = float(cal_dict[det]["pz"]["tau1"][:-3]) / 1000
            except KeyError:
                taus[det] = np.nan

    p = figure(
        width=1400,
        height=600,
        y_range=(350, 800),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | Pole Zero Constants"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(taus)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    pz_types = ["pz_constant"]
    pz_names = ["PZ constant"]
    pz_colors = ["blue"]

    for pz_type in pz_types:
        x_plot, y_plot, y_plot_err = (
            np.arange(1, len(list(taus)) + 1, 1),
            [taus[det] for det in taus],
            [10] * len(taus),
        )

        err_xs = []
        err_ys = []

        for x, y, yerr in zip(x_plot, y_plot, y_plot_err, strict=False):
            err_xs.append((x, x))
            err_ys.append((np.nan_to_num(y - yerr), np.nan_to_num(y + yerr)))

        df_plot["x_{}".format(pz_type.split("_")[0])] = np.nan_to_num(x_plot)
        df_plot["y_{}".format(pz_type.split("_")[0])] = np.nan_to_num(y_plot)
        df_plot["y_{}_err".format(pz_type.split("_")[0])] = np.nan_to_num(y_plot_err)
        df_plot["err_xs_{}".format(pz_type.split("_")[0])] = err_xs
        df_plot["err_ys_{}".format(pz_type.split("_")[0])] = err_ys

    if download:
        return (
            df_plot,
            f"{run_dict['experiment']}-{period}-{run}_PoleZero_Constants.csv",
        )

    # df_plot = ColumnDataSource(df_plot)
    for pz_type, pz_name, pz_color in zip(pz_types, pz_names, pz_colors, strict=False):
        if pz_type == "pz_constant":
            hover_renderer = p.scatter(
                x="x_{}".format(pz_type.split("_")[0]),
                y="y_{}".format(pz_type.split("_")[0]),
                source=df_plot,
                color=pz_color,
                size=7,
                line_alpha=0,
                legend_label=pz_name,
                name=pz_name,
            )
        else:
            p.scatter(
                x="x_{}".format(pz_type.split("_")[0]),
                y="y_{}".format(pz_type.split("_")[0]),
                source=df_plot,
                color=pz_color,
                size=7,
                line_alpha=0,
                legend_label=pz_name,
                name=pz_name,
            )
        p.multi_line(
            xs="err_xs_{}".format(pz_type.split("_")[0]),
            ys="err_ys_{}".format(pz_type.split("_")[0]),
            source=df_plot,
            color=pz_color,
            legend_label=pz_name,
            name=pz_name,
        )

    p.legend.location = "top_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "PZ constant (µs)"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(taus)), 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(taus)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=350, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        ("PZ const.", "@y_pz{0.0} +- @y_pz_err{0.0} µs"),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def plot_alpha(
    prod_config,
    run,
    run_dict,
    path,
    period,
    key="String",
    sort_dets_obj=None,
    download=False,
    cache_data=None,
):
    strings, soft_dict, channel_map = sorter(
        path, run_dict["timestamp"], key=key, sort_dets_obj=sort_dets_obj
    )

    if cache_data is not None and run in cache_data["dsp"]:
        cal_dict = cache_data["dsp"][run]
    else:
        file_path = Path(prod_config["paths"]["par_dsp"]) / f"cal/{period}/{run}"
        path = (
            file_path
            / f'{run_dict["experiment"]}-{period}-{run}-cal-{run_dict["timestamp"]}-par_dsp.yaml'
        )

        cal_dict = Props.read_from(path)
        cache_data["dsp"][run] = cal_dict

    trap_alpha = {}
    cusp_alpha = {}
    zac_alpha = {}

    for stri in strings:
        trap_alpha[stri] = np.nan
        cusp_alpha[stri] = np.nan
        zac_alpha[stri] = np.nan
        for channel in strings[stri]:
            det = channel_map[channel]["name"]
            try:
                trap_alpha[det] = float(
                    cal_dict[det]["ctc_params"]["trapEmax_ctc"]["parameters"]["a"]
                )
                cusp_alpha[det] = float(
                    cal_dict[det]["ctc_params"]["cuspEmax_ctc"]["parameters"]["a"]
                )
                zac_alpha[det] = float(
                    cal_dict[det]["ctc_params"]["zacEmax_ctc"]["parameters"]["a"]
                )
            except KeyError:
                trap_alpha[det] = np.nan
                cusp_alpha[det] = np.nan
                zac_alpha[det] = np.nan

    p = figure(
        width=1400,
        height=600,
        y_range=(-1, 4),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | Charge Trapping Constants"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(cusp_alpha)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    df_plot["x_cuspEmax_ctc_cal"] = np.arange(1, len(list(cusp_alpha)) + 1, 1)
    df_plot["x_zacEmax_ctc_cal"] = np.arange(1, len(list(zac_alpha)) + 1, 1)
    df_plot["x_trapEmax_ctc_cal"] = np.arange(1, len(list(trap_alpha)) + 1, 1)
    df_plot["cuspEmax_ctc_cal"] = np.nan_to_num(list(cusp_alpha.values())) * 1e6
    df_plot["zacEmax_ctc_cal"] = np.nan_to_num(list(zac_alpha.values())) * 1e6
    df_plot["trapEmax_ctc_cal"] = np.nan_to_num(list(trap_alpha.values())) * 1e6

    filter_types = ["cuspEmax_ctc_cal", "zacEmax_ctc_cal", "trapEmax_ctc_cal"]
    filter_names = ["Cusp", "ZAC", "Trap"]
    filter_plot_colors = ["blue", "green", "red"]

    if download:
        return (
            df_plot,
            f"{run_dict['experiment']}-{period}-{run}_Charge_Trapping_Constants.csv",
        )

    # df_plot = ColumnDataSource(df_plot)
    for filter_type, filter_name, filter_plot_color in zip(
        filter_types, filter_names, filter_plot_colors, strict=False
    ):
        if filter_name == "Cusp":
            hover_renderer = p.scatter(
                x=f"x_{filter_type}",
                y=filter_type,
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=filter_name,
                name=filter_name,
            )
        else:
            p.scatter(
                x=f"x_{filter_type}",
                y=filter_type,
                source=df_plot,
                color=filter_plot_color,
                size=7,
                line_alpha=0,
                legend_label=filter_name,
                name=filter_name,
            )

    p.legend.location = "top_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = r"$$\text{Alpha Value} (\frac{10^{-6}}{\text{ns}})$$"
    p.yaxis.axis_label_text_font_size = "16px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(cusp_alpha)), 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(cusp_alpha)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=-0.95, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        ("Alpha Cusp", "@cuspEmax_ctc_cal{0.00e}"),
        ("Alpha ZAC ", "@zacEmax_ctc_cal{0.00}"),
        ("Alpha Trap", "@trapEmax_ctc_cal{0.00}"),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p


def plot_bls(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    cache_data=None,
):
    p = figure(
        width=700,
        height=600,
        y_axis_type="log",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | Baseline | {string}"
    )
    p.title.align = "center"
    p.title.text_font_size = "15px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(channels)
    colours = cc.palette["glasbey_category10"][:len_colours]

    for i, channel in enumerate(channels):
        try:
            plot_dict_chan = plot_dict[channel]

            p.step(
                plot_dict_chan["baseline_spectrum"]["bins"],
                plot_dict_chan["baseline_spectrum"]["bl_array"],
                legend_label=f'{chan_dict[channel]["name"]}',
                mode="after",
                name=f'{chan_dict[channel]["name"]}',
                line_width=2,
                line_color=colours[i],
            )
        except KeyError:
            pass

    p.hover.tooltips = [("Detector", "$name"), ("Baseline", "$x"), ("Counts", "$y")]
    p.hover.mode = "vline"
    p.xaxis.axis_label = "Wf Baseline Mean - FC Baseline"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Counts"
    p.yaxis.axis_label_text_font_size = "16px"
    p.legend.location = "top_left"
    p.legend.click_policy = "hide"
    return p


def plot_energy_spectra(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    energy_param="cuspEmax_ctc_cal",
    cache_data=None,
):
    p = figure(
        width=700,
        height=600,
        y_axis_type="log",
        x_axis_type="datetime",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Cal. | Energy Spectra | {string}"
    )
    p.title.align = "center"
    p.title.text_font_size = "15px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(channels)
    colours = cc.palette["glasbey_category10"][:len_colours]

    for i, channel in enumerate(channels):
        try:
            plot_dict_chan = plot_dict[channel]
            p.step(
                plot_dict_chan[energy_param]["spectrum"]["bins"],
                plot_dict_chan[energy_param]["spectrum"]["counts"],
                legend_label=f'{chan_dict[channel]["name"]}',
                mode="after",
                name=f'{chan_dict[channel]["name"]}',
                line_width=2,
                line_color=colours[i],
            )
        except KeyError:
            pass

    p.xaxis.axis_label = "Energy (keV)"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Counts"
    p.yaxis.axis_label_text_font_size = "16px"
    p.legend.location = "top_left"
    p.legend.click_policy = "hide"

    return p


def plot_baseline_stability(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    cache_data=None,
):
    times = None
    p = figure(
        width=700,
        height=600,
        x_axis_type="datetime",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | Baseline Stability | {string}"
    p.title.align = "center"
    p.title.text_font_size = "15px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(channels)
    colours = cc.palette["glasbey_category10"][:len_colours]

    for i, channel in enumerate(channels):
        try:
            bl = plot_dict[channel]["baseline_stability"]["baseline"]
            # bl_spread = plot_dict[channel]["baseline_stability"]["spread"]
            mean = np.nanmean(bl[~np.isnan(bl)][:10])
            bl_mean = 100 * (bl - mean) / mean

            # define if condition such that timedelta only added if still in UTC
            base_time = plot_dict[channel]["baseline_stability"]["time"][0]
            dt_object_base = datetime.utcfromtimestamp(base_time)
            utc_offset_base = dt_object_base.utcoffset()

            if utc_offset_base is None:
                p.line(
                    [
                        (datetime.fromtimestamp(time) + timedelta(hours=2))
                        for time in plot_dict[channel]["baseline_stability"]["time"]
                    ],  # add two hours manually
                    bl_mean,
                    legend_label=f'{chan_dict[channel]["name"]}',
                    name=f'{chan_dict[channel]["name"]}',
                    line_width=2,
                    line_color=colours[i],
                )
                if times is None:
                    times = [
                        (datetime.fromtimestamp(t) + timedelta(hours=2))
                        for t in plot_dict[channel]["baseline_stability"]["time"]
                    ]
            if utc_offset_base is not None:
                p.line(
                    [
                        datetime.fromtimestamp(time)
                        for time in plot_dict[channel]["baseline_stability"]["time"]
                    ],
                    bl_mean,
                    legend_label=f'{chan_dict[channel]["name"]}',
                    name=f'{chan_dict[channel]["name"]}',
                    line_width=2,
                    line_color=colours[i],
                )
                if times is None:
                    times = [
                        datetime.fromtimestamp(t)
                        for t in plot_dict[f"ch{channel:03}"]["baseline_stability"][
                            "time"
                        ]
                    ]
        except KeyError:
            pass

    # revision of hover tool to display values correctly
    p.hover.formatters = {"$x": "datetime", "$y": "printf"}
    p.hover.tooltips = [
        ("Detector", "$name"),
        ("Time", "$x{%F %H:%M:%S CET}"),
        ("BL Shift (%)", "@y{0, 0.0000} %"),
    ]
    p.hover.mode = "vline"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {times[0].strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Shift (%)"
    p.yaxis.axis_label_text_font_size = "16px"
    p.legend.location = "top_left"
    p.legend.click_policy = "hide"
    return p


def plot_stability(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    parameter,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    energy_param="cuspEmax_ctc",
    cache_data=None,
):
    times = None
    p = figure(
        width=700,
        height=600,
        x_axis_type="datetime",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    if parameter == "2614_stability":
        p.title.text = (
            f"{run_dict['experiment']}-{period}-{run} | Cal. | FEP Stability | {string}"
        )
    else:
        p.title.text = f"{run_dict['experiment']}-{period}-{run} | Cal. | Pulser Stability | {string}"
    p.title.align = "center"
    p.title.text_font_size = "15px"

    len_colours = len(channels)
    colours = cc.palette["glasbey_category10"][:len_colours]

    for i, channel in enumerate(channels):
        try:
            plot_dict_chan = plot_dict[channel]

            en = plot_dict_chan[energy_param][parameter]["energy"]
            # en_spread = plot_dict_chan[energy_param][parameter]["spread"]
            mean = np.nanmean(en[~np.isnan(en)][:10])
            en_mean = en - mean  # /mean

            # define if condition such that timedelta only added if still in UTC
            plot_time = plot_dict_chan[energy_param][parameter]["time"][0]
            dt_object_plot = datetime.utcfromtimestamp(plot_time)
            utc_offset = dt_object_plot.utcoffset()

            if utc_offset is None:
                p.line(
                    [
                        (datetime.fromtimestamp(time) + timedelta(hours=2))
                        for time in plot_dict_chan[energy_param][parameter]["time"]
                    ],  # add two hours manually
                    en_mean,
                    legend_label=f'{chan_dict[channel]["name"]}',
                    name=f'{chan_dict[channel]["name"]}',
                    line_width=2,
                    line_color=colours[i],
                )
                if times is None:
                    times = [
                        (datetime.fromtimestamp(t) + timedelta(hours=2))
                        for t in plot_dict_chan[energy_param][parameter]["time"]
                    ]
            if utc_offset is not None:
                p.line(
                    [
                        (datetime.fromtimestamp(time))
                        for time in plot_dict_chan[energy_param][parameter]["time"]
                    ],
                    en_mean,
                    legend_label=f'{chan_dict[channel]["name"]}',
                    name=f'{chan_dict[channel]["name"]}',
                    line_width=2,
                    line_color=colours[i],
                )
                if times is None:
                    times = [
                        datetime.fromtimestamp(t)
                        for t in plot_dict_chan[energy_param][parameter]["time"]
                    ]
        except KeyError:
            pass

    # revision of hover tool to display detector name and values correctly
    p.hover.formatters = {"$x": "datetime", "$y": "printf"}
    p.hover.tooltips = [
        ("Detector", "$name"),
        ("Time", "$x{%F %H:%M:%S CET}"),
        ("Energy Shift (%)", "@y{0, 0.0000} %"),
    ]
    p.hover.mode = "vline"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {times[0].strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Energy Shift (keV)"
    p.yaxis.axis_label_text_font_size = "16px"
    p.legend.location = "top_left"
    p.legend.click_policy = "hide"
    return p


def plot_fep_stability_channels2d(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    energy_param="cuspEmax_ctc_cal",
    cache_data=None,
):
    return plot_stability(
        prod_config,
        plot_dict,
        chan_dict,
        channels,
        string,
        "2614_stability",
        run,
        period,
        run_dict,
        key=key,
        sort_dets_obj=sort_dets_obj,
        energy_param=energy_param,
        cache_data=cache_data,
    )


def plot_pulser_stability_channels2d(
    prod_config,
    plot_dict,
    chan_dict,
    channels,
    string,
    run,
    period,
    run_dict,
    key="String",
    sort_dets_obj=None,
    energy_param="cuspEmax_ctc_cal",
    cache_data=None,
):
    return plot_stability(
        prod_config,
        plot_dict,
        chan_dict,
        channels,
        string,
        "pulser_stability",
        run,
        period,
        run_dict,
        key=key,
        sort_dets_obj=sort_dets_obj,
        energy_param=energy_param,
        cache_data=cache_data,
    )



from __future__ import annotations

from datetime import UTC, datetime
from pathlib import Path

import importlib.resources

import matplotlib as mpl
import numpy as np
import panel as pn
from dbetto import AttrsDict, Props, TextDB
from dbetto.catalog import Catalog
from legendmeta import LegendMetadata

# somehow TUM server needs Agg -> needs fix in the future
mpl.use("Agg")

# use terminal and tabulator extensions, sizing mode stretch enables nicer layout
pn.extension("terminal")
pn.extension("tabulator")
pn.extension("plotly")
pn.extension("katex", "mathjax")

logo_path = importlib.resources.files("legenddashboard") / "logos"

sort_dict = {
    "String": {
        "out_key": "{key}:{k:02}",
        "primary_key": "location.string",
        "secondary_key": "location.position",
    },
    "CC4": {
        "out_key": "{key}:{k}",
        "primary_key": "electronics.cc4.id",
        "secondary_key": "electronics.cc4.channel",
    },
    "HV": {
        "out_key": "{key}:{k:02}",
        "primary_key": "voltage.card.id",
        "secondary_key": "voltage.channel",
    },
    "Det_Type": {"out_key": "{k}", "primary_key": "type", "secondary_key": "name"},
    "DAQ": {"out_key": None, "primary_key": None, "secondary_key": None},
}


def read_config(config: str | dict) -> AttrsDict:
    """
    Parse the config file or dictionary and return an AttrsDict.
    """
    if isinstance(config, str | Path):
        config = AttrsDict(Props.read_from(config))
    else:
        config = AttrsDict(config)

    return config.paths


class sort_dets:
    def __init__(self, path):
        self.cached_chmaps = {}
        self.cached_det_status = {}

        path = Path(path)

        self.prod_config = path / "dataflow-config.yaml"
        self.prod_config = Props.read_from(self.prod_config, subst_pathvar=True)

        meta_path = Path(self.prod_config["paths"]["metadata"])
        chmap_path = Path(self.prod_config["paths"]["chan_map"])
        chmap_catalog = Catalog.read_from(chmap_path / "channelmaps" / "validity.yaml")
        chmap_entries = {}
        meta = LegendMetadata(meta_path, lazy=False)
        for system in chmap_catalog.entries:
            chmap_entries[system] = []
            for entry in chmap_catalog.entries[system]:
                try:
                    db = meta.channelmap(
                        datetime.fromtimestamp(entry.valid_from, tz=UTC), system=system
                    )
                    new_entry = Catalog.Entry(entry.valid_from, db)
                    chmap_entries[system].append(new_entry)
                except RuntimeError:
                    continue

        self.chmaps = Catalog(chmap_entries)

        status_path = Path(self.prod_config["paths"]["detector_status"]) / "statuses"
        status_catalog = Catalog.read_from(status_path / "validity.yaml")
        status_entries = {}
        textdb = TextDB(status_path, lazy=False)
        for system in status_catalog.entries:
            status_entries[system] = []
            for entry in status_catalog.entries[system]:
                db = textdb.on(
                    datetime.fromtimestamp(entry.valid_from, tz=UTC), system=system
                )
                new_entry = Catalog.Entry(entry.valid_from, db)
                status_entries[system].append(new_entry)

        self.statuses = Catalog(status_entries)


def gen_run_dict(path):
    prod_config = Path(path) / "dataflow-config.yaml"
    prod_config = Props.read_from(prod_config, subst_pathvar=True)  # ["setups"]["l200"]
    par_file = Path(prod_config["paths"]["par_hit"]) / "validity.yaml"
    run_dict = {}
    file = Props.read_from(par_file)
    for entry in file:
        experiment, period, run, _, _, _ = entry["apply"][0].split("/")[-1].split("-")
        timestamp = entry["valid_from"]
        if (
            Path(prod_config["paths"]["par_hit"])
            / f"cal/{period}/{run}"
            / f"{experiment}-{period}-{run}-cal-{timestamp}-par_hit.yaml"
        ).exists():
            if period in run_dict:
                run_dict[period][run] = {
                    "experiment": experiment,
                    "timestamp": timestamp,
                }
            else:
                run_dict[period] = {
                    run: {"experiment": experiment, "timestamp": timestamp}
                }
    return run_dict


def sorter(
    path, timestamp, key="String", datatype="cal", spms=False, sort_dets_obj=None
):
    if sort_dets_obj is not None:
        chmap = sort_dets_obj.chmaps.valid_for(timestamp, system=datatype)
        det_status = sort_dets_obj.statuses.valid_for(timestamp, system=datatype)
    else:
        prod_config = Path(path) / "dataflow-config.yaml"
        prod_config = Props.read_from(
            prod_config, subst_pathvar=True
        )  # ["setups"]["l200"]

        cfg_file = prod_config["paths"]["metadata"]
        configs = LegendMetadata(path=cfg_file)
        chmap = configs.channelmap(timestamp)

        det_status_path = prod_config["paths"]["detector_status"]
        det_status = LegendMetadata(path=det_status_path, lazy=True).statuses.on(
            timestamp, system=datatype
        )

    out_dict = {}
    # SiPMs sorting
    if spms:
        chmap = chmap.map("system", unique=False)["spms"]
        if key == "Barrel":
            mapping = chmap.map("name")
            for pos in ["top", "bottom"]:
                for barrel in ["IB", "OB"]:
                    out_dict[f"{barrel}-{pos}"] = [
                        k
                        for k, entry in sorted(mapping.items())
                        if barrel in entry["location"]["fiber"]
                        and pos in entry["location"]["position"]
                    ]
        return out_dict, chmap

    # Daq needs special item as sort on tertiary key
    if key == "DAQ":
        mapping = chmap.map("daq.crate", unique=False)
        for k, entry in sorted(mapping.items()):
            for m, item in sorted(entry.map("daq.card.id", unique=False).items()):
                out_dict[f"DAQ:Cr{k:02},Ch{m:02}"] = [
                    item.map("daq.channel")[pos].name
                    for pos in sorted(item.map("daq.channel"))
                    if item.map("daq.channel")[pos].system == "geds"
                ]
    else:
        out_key = sort_dict[key]["out_key"]
        primary_key = sort_dict[key]["primary_key"]
        secondary_key = sort_dict[key]["secondary_key"]
        mapping = chmap.map(primary_key, unique=False)
        for k, entry in sorted(mapping.items()):
            out_dict[out_key.format(key=key, k=k)] = [
                entry.map(secondary_key)[pos].name
                for pos in sorted(entry.map(secondary_key))
                if entry.map(secondary_key)[pos].system == "geds"
            ]

    out_dict = {
        entry: out_dict[entry] for entry in list(out_dict) if len(out_dict[entry]) > 0
    }
    return out_dict, det_status, chmap


def get_characterization(x, key):
    try:
        return x["manufacturer"][key]
    except KeyError:
        return np.nan


def get_production(x, key):
    try:
        return x[key]
    except KeyError:
        return np.nan



from __future__ import annotations

import argparse
import logging
import time
from pathlib import Path

import pandas as pd
import panel as pn
import param
from bokeh.models.formatters import PrintfTickFormatter
from bokeh.plotting import figure

import legenddashboard.spms.sipm_plots as spm
from legenddashboard.base import Monitoring
from legenddashboard.util import logo_path, read_config, sorter

log = logging.getLogger(__name__)

sipm_plot_style_dict = {
    "Time": spm.sipm_plot_vsTime,
    "Histogram": spm.sipm_plot_histogram,
}


class SiPMMonitoring(Monitoring):
    sipm_path = param.String("")

    # sipm plots
    # sipm_plots_barrels    = ['InnerBarrel-Top', 'InnerBarrel-Bottom', 'OuterBarrel-Top', 'OuterBarrel-Bottom']
    # sipm_resampled_vals = [1, 5, 10, 30, 60]

    sipm_sort_by = param.ObjectSelector(objects=["Barrel"], default="Barrel")

    sipm_barrel = param.ObjectSelector(default=0, objects=[0])
    sipm_resampled = param.Integer(
        default=1,
        bounds=(1, 60),
    )
    sipm_plot_style = param.ObjectSelector(
        default=next(iter(sipm_plot_style_dict)),
        objects=list(sipm_plot_style_dict),
    )
    sipm_plot_style_dict = param.Dict(sipm_plot_style_dict)

    # self._get_sipm_data()

    @param.depends("sipm_sort_by", watch=True)
    def update_barrels(self):
        start_time = time.time()
        self.sipm_out_dict, self.sipm_chmap = sorter(
            self.path,
            self.run_dict[self.run]["timestamp"],
            key=self.sipm_sort_by,
            spms=True,
            sort_dets_obj=self.sort_obj,
        )

        self.param["sipm_barrel"].objects = list(self.sipm_out_dict)
        self.sipm_barrel = f"{next(iter(self.sipm_out_dict))}"

        log.debug("Time to update barrels:", extra={"time": time.time() - start_time})

    @param.depends("run", watch=True)
    def _get_sipm_data(self):
        start_time = time.time()
        data_file = self.sipm_path + f"{self.period}_{self.run}_spmmon.hdf"
        if not Path(data_file).exists():
            self.sipm_data_df = pd.DataFrame()
        else:
            self.sipm_data_df = (
                pd.read_hdf(data_file)
                .reset_index()
                .set_index("time")
                .drop(["index"], axis=1)
            )
            self.sipm_data_df.index = pd.to_datetime(
                self.sipm_data_df.index, unit="s", origin="unix"
            )

        self.sipm_out_dict, self.sipm_chmap = sorter(
            self.path,
            self.run_dict[self.run]["timestamp"],
            key=self.sipm_sort_by,
            spms=True,
            sort_dets_obj=self.sort_obj,
        )
        self.sipm_name_dict = {}
        for val in self.sipm_chmap.values():
            self.sipm_name_dict[val["daq"]["rawid"]] = val["name"]
        self.update_barrels()
        log.debug("Time to get sipm data:", extra={"time": time.time() - start_time})

    @param.depends(
        "run", "sipm_sort_by", "sipm_resampled", "sipm_barrel", "sipm_plot_style"
    )
    def view_sipm(self):
        start_time = time.time()
        if self.sipm_data_df.empty:
            p = figure(width=1000, height=600)
            p.title.text = f"No data for run {self.run_dict[self.run]['experiment']}-{self.period}-{self.run}"
            p.title.align = "center"
            p.title.text_font_size = "25px"
            log.debug(
                "Time to get sipm plot:", extra={"time": time.time() - start_time}
            )
            return p
        data_barrel = self.sipm_data_df[
            [
                f"ch{channel}"
                for channel in self.sipm_out_dict[self.sipm_barrel]
                if f"ch{channel}" in self.sipm_data_df.columns
            ]
        ]
        p = self.sipm_plot_style_dict[self.sipm_plot_style](
            data_barrel,
            self.sipm_barrel,
            f"{self.sipm_resampled}min",
            self.sipm_name_dict,
            self.run,
            self.period,
            self.run_dict[self.run],
        )
        log.debug("Time to get sipm plot:", extra={"time": time.time() - start_time})
        return p

    def build_spm_pane(
        self,
        widget_widths=140,
    ):
        sipm_param_currentValue = pn.pane.Markdown(f"## {self.sipm_barrel}")
        sipm_param = pn.widgets.MenuButton(
            name="SiPM",
            button_type="primary",
            width=widget_widths,
            items=self.param.sipm_barrel.objects,
        )

        def update_sipm_plots(event):
            self.sipm_barrel = event.new
            sipm_param_currentValue.object = f"## {event.new}"

        sipm_param.on_click(update_sipm_plots)

        sipm_plot_style_param = pn.Param(
            self.param,
            widgets={
                "sipm_plot_style": {
                    "widget_type": pn.widgets.RadioButtonGroup,
                    "button_type": "light",
                    "orientation": "vertical",
                    "width": widget_widths,
                }
            },
            parameters=["sipm_plot_style"],
            show_labels=False,
            show_name=False,
            sort=False,
        )

        sipm_plot_param = pn.Param(
            self.param,
            widgets={
                "sipm_resampled": {
                    "widget_type": pn.widgets.IntSlider,
                    "width": widget_widths,
                    "format": PrintfTickFormatter(format="%d min"),
                }
            },
            parameters=["sipm_resampled"],
            show_labels=False,
            show_name=False,
            sort=False,
        )

        return pn.Column(
            pn.Row(pn.pane.SVG(logo_path / "Physics.svg", height=25), sipm_param),
            pn.Row("## Current Plot:", sipm_param_currentValue),
            pn.Row(sipm_plot_style_param, pn.Column("Resampled", sipm_plot_param)),
            pn.pane.Bokeh(self.view_sipm, sizing_mode="scale_both"),
            name="SiPM Monitoring",
            sizing_mode="scale_both",
        )

    @classmethod
    def init_spm_pane(
        cls,
        base_path,
        spm_path,
        widget_widths=140,
    ):
        """
        Initialize the SiPM monitoring pane.
        """
        return cls(
            path=base_path,
            sipm_path=spm_path,
        ).build_spm_pane(widget_widths)

    @classmethod
    def display_spms(
        cls,
        base_path,
        notebook=False,
        widget_widths: int = 140,
    ):
        """
        View the SiPM panes.

        Args:
            widget_widths (int): Width of the widgets.

        Returns:
            pn.Row: Row containing the sidebar and the spm pane.
        """
        spm_monitor = cls(base_path=base_path, notebook=notebook)
        sidebar = spm_monitor.build_sidebar()
        return pn.Row(sidebar, spm_monitor.build_spm_pane(widget_widths))


def run_dashboard_spms() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    spm_pane = SiPMMonitoring.display_spms(config.base, args.widget_widths)
    print("Starting SiPM Monitoring on port ", args.port)  # noqa: T201
    pn.serve(spm_pane, port=args.port, show=False)



from __future__ import annotations

import argparse
import logging
import time
from pathlib import Path

import h5py
import pandas as pd
import panel as pn
import param
from bokeh.models.formatters import PrintfTickFormatter
from bokeh.plotting import figure

from legenddashboard.geds import phy
from legenddashboard.geds.ged_monitoring import GedMonitoring
from legenddashboard.util import logo_path, read_config

log = logging.getLogger(__name__)


class PhyMonitoring(GedMonitoring):
    phy_path = param.String("")
    phy_plots_types = param.ObjectSelector(
        default=next(iter(phy.phy_plots_types_dict)),
        objects=list(phy.phy_plots_types_dict),
        label="Type",
    )
    phy_plots = param.ObjectSelector(
        default=list(phy.phy_plots_vals_dict)[4],
        objects=list(phy.phy_plots_vals_dict),
        label="Value",
    )
    phy_plot_style = param.ObjectSelector(
        default=next(iter(phy.phy_plot_style_dict)),
        objects=list(phy.phy_plot_style_dict),
        label="Plot Style",
    )
    phy_resampled = param.Integer(
        default=phy.phy_resampled_vals[0],
        bounds=(phy.phy_resampled_vals[0], phy.phy_resampled_vals[-1]),
    )
    phy_units = param.ObjectSelector(
        default=phy.phy_unit_vals[0], objects=phy.phy_unit_vals, label="Units"
    )
    # phy_plots_sc        = param.Boolean(default=False, label="SC")
    phy_plots_sc_vals = param.ObjectSelector(
        default=next(iter(phy.phy_plots_sc_vals_dict)),
        objects=list(phy.phy_plots_sc_vals_dict),
        label="SC Values",
    )

    # create initial dataframes
    phy_data_df = pd.DataFrame()
    phy_data_df_mean = pd.DataFrame()
    phy_abs_unit = ""
    phy_plot_info = None
    phy_data_sc = pd.DataFrame()
    phy_pane = pn.pane.Bokeh(figure(width=1000, height=600), sizing_mode="scale_width")
    _phy_sc_plotted = False

    @param.depends(
        "run",
        "string",
        "sort_by",
        "phy_plots_types",
        "phy_plots",
        "phy_plot_style",
        "phy_resampled",
        "phy_units",
        "phy_plots_sc_vals",
    )
    def update_plots(self):
        start_time = time.time()
        data_file = (
            self.phy_path
            + f"/generated/plt/phy/{self.period}/{self.run}/l200-{self.period}-{self.run}-phy-geds.hdf"
        )
        data_file_sc = (
            self.phy_path
            + f"/generated/plt/phy/{self.period}/{self.run}/l200-{self.period}-{self.run}-phy-slow_control.hdf"
        )

        # Create empty plot inc ase of errors
        p = figure(width=1000, height=600)
        p.title.text = f"No data for run {self.run_dict[self.run]['experiment']}-{self.period}-{self.run}"
        p.title.align = "center"
        p.title.text_font_size = "25px"

        # return empty plot if no data exists for run
        if not Path(data_file).exists():
            log.debug("Time to get phy plot:", extra={"time": time.time() - start_time})
            return p

        # get filekeys to check if key exists
        with h5py.File(data_file, "r") as f:
            filekeys = list(f.keys())

        # load dataframe for current plot value and get all data from selected string
        channels = self.strings_dict[self.string]
        phy_data_key = f"{self.phy_plots_types_dict[self.phy_plots_types]}_{self.phy_plots_vals_dict[self.phy_plots]}"
        if "pulser" in phy_data_key:
            if f"{phy_data_key.split('_pulser')[0]}_info" not in filekeys:
                return p
            phy_plot_info = pd.read_hdf(
                data_file, key=f"{phy_data_key.split('_pulser')[0]}_info"
            )
            if "Diff" in phy_data_key:
                phy_plot_info.loc["label"][0] = "Gain to Pulser Difference"
            else:
                phy_plot_info.loc["label"][0] = "Gain to Pulser Ratio"
        else:
            if f"{phy_data_key}_info" not in filekeys:
                return p
            phy_plot_info = pd.read_hdf(data_file, key=f"{phy_data_key}_info")
        abs_unit = phy_plot_info.loc["unit"][0]

        if self.phy_units == "Relative":
            if f"{phy_data_key}_var" not in filekeys:
                return p
            phy_data_df = pd.read_hdf(data_file, key=f"{phy_data_key}_var")
            phy_plot_info.loc["unit"][0] = "%"
        else:
            if phy_data_key not in filekeys:
                return p
            phy_data_df = pd.read_hdf(data_file, key=phy_data_key)

        # load mean values
        if f"{phy_data_key}_mean" not in filekeys:
            return p
        phy_data_df_mean = pd.read_hdf(data_file, key=f"{phy_data_key}_mean")

        # get sc data if selected
        # if self.phy_plots_sc and self.phy_units == "Relative" and os.path.exists(data_file_sc):
        if (
            self.phy_plots_sc_vals_dict[self.phy_plots_sc_vals]
            and Path(data_file_sc).exists()
        ):
            data_sc = pd.read_hdf(
                data_file_sc, self.phy_plots_sc_vals_dict[self.phy_plots_sc_vals]
            )
            self._phy_sc_plotted = True
        else:
            data_sc = pd.DataFrame()
            self._phy_sc_plotted = False
        # check if channel selection actually exists in data
        channels = [
            ch
            for ch in channels
            if ch in phy_data_df.columns and ch in phy_data_df_mean.columns
        ]
        phy_data_df = phy_data_df[channels]
        phy_data_df_mean = phy_data_df_mean[channels]

        # plot data
        p = self.phy_plot_style_dict[self.phy_plot_style](
            phy_data_df,
            phy_data_df_mean,
            phy_plot_info,
            self.phy_plots_types,
            self.phy_plots,
            f"{self.phy_resampled}min",
            self.string,
            self.run,
            self.period,
            self.run_dict[self.run],
            self.channel_map,
            abs_unit,
            data_sc,
            self.phy_plots_sc_vals,
        )
        log.debug("Time to get phy plot:", extra={"time": time.time() - start_time})
        # self.bokeh_pane.object = p
        return p

    def build_phy_pane(self, widget_widths=140):
        """
        Build the phy pane with all widgets and plots
        """

        # physics_style_param = pn.Param(
        #     self.param,
        #     widgets={
        #         "phy_plot_style": {
        #             "widget_type": pn.widgets.RadioButtonGroup,
        #             "button_type": "light",
        #             "orientation": "vertical",
        #             "width": widget_widths,
        #         }
        #     },
        #     parameters=["phy_plot_style"],
        #     show_labels=False,
        #     show_name=False,
        #     sort=False,
        # )
        physics_param_resampled = pn.Param(
            self.param,
            widgets={
                "phy_resampled": {
                    "widget_type": pn.widgets.IntSlider,
                    "width": widget_widths,
                    "format": PrintfTickFormatter(format="%d min"),
                    "value_throttled": True,
                },
            },
            parameters=["phy_resampled"],
            show_labels=False,
            show_name=False,
            sort=False,
        )
        physics_param_units = pn.Param(
            self.param,
            widgets={
                "phy_units": {"widget_type": pn.widgets.RadioBoxGroup, "inline": True}
            },
            parameters=["phy_units"],
            show_labels=False,
            show_name=False,
            sort=False,
        )
        physics_param_types = pn.Param(
            self.param,
            widgets={
                "phy_plots_types": {
                    "widget_type": pn.widgets.RadioButtonGroup,
                    "orientation": "vertical",
                    "button_type": "primary",
                    "button_style": "outline",
                    "width": widget_widths,
                }
            },
            parameters=["phy_plots_types"],
            show_labels=False,
            show_name=False,
            sort=False,
        )

        physics_param_currentValue = pn.pane.Markdown(f"## {self.phy_plots}")
        physics_param = pn.widgets.MenuButton(
            name="HPGe Detectors",
            button_type="primary",
            width=widget_widths,
            items=self.param.phy_plots.objects,
        )

        def update_phy_plots(event):
            self.phy_plots = event.new
            physics_param_currentValue.object = f"## {event.new}"

        physics_param.on_click(update_phy_plots)

        # SC
        # sc_param_currentValue = pn.pane.Markdown(f"## Not selected or no data available")
        sc_param = pn.widgets.MenuButton(
            name="Slow Control",
            button_type="warning",
            width=widget_widths,
            items=self.param.phy_plots_sc_vals.objects,
        )

        def update_sc_plots(event):
            self.phy_plots_sc_vals = event.new
            # if self.phy_plots_sc_vals == "None" or not self._phy_sc_plotted:
            #         sc_param_currentValue.object = f"## Not selected or no data available"
            # else:
            #         sc_param_currentValue.object = f"## {event.new}"

        sc_param.on_click(update_sc_plots)

        phy_gspec = pn.GridSpec(width=3 * widget_widths + 10, max_height=800)
        phy_gspec[:, 0] = physics_param_types
        phy_gspec[:, 1] = pn.Spacer(width=5)
        phy_gspec[0, 2] = pn.widgets.Button(
            name="Units", button_type="primary", width=widget_widths, disabled=True
        )
        phy_gspec[1, 2] = physics_param_units
        phy_gspec[:, 3] = pn.Spacer(width=5)
        phy_gspec[0, 4] = pn.widgets.Button(
            name="Resampled", button_type="primary", width=widget_widths, disabled=True
        )
        phy_gspec[1, 4] = physics_param_resampled
        # phy_gspec[:, 5] = pn.Spacer(width=5)
        # phy_gspec[0, 6] = pn.widgets.Button(name="Show Slow Control", button_type='danger', width=widget_widths, disabled=True)
        # phy_gspec[1, 6] = sc_param_selected
        # pn.Row(physics_param_types, pn.Column("Units", physics_param_units), pn.Column("Resampled", physics_param_resampled), pn.Column("Show Slow Control", sc_param_selected)),
        # pn.Row(phy_gspec)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Physics.svg",
                    height=25,
                ),
                physics_param,
                sc_param,
            ),
            pn.Row("## Current Plot:", physics_param_currentValue),
            # pn.Row("## Current SC Plot:", sc_param_currentValue),
            pn.Row(phy_gspec),
            pn.pane.Bokeh(self.update_plots(), sizing_mode="scale_width"),
            name="Phy. Monitoring",
            sizing_mode="stretch_width",
        )

    @classmethod
    def init_phy_panes(
        cls,
        base_path,
        phy_path,
        widget_widths: int = 140,
    ):
        phy_monitor = cls(
            base_path=base_path,
            phy_path=phy_path,
        )
        return phy_monitor.build_phy_pane(widget_widths)

    @classmethod
    def display_phy(
        cls,
        base_path,
        notebook=False,
        widget_widths: int = 140,
    ):
        """
        View the Physics panes.

        Args:
            widget_widths (int): Width of the widgets.

        Returns:
            pn.Row: Row containing the sidebar and the phy pane.
        """
        phy_monitor = cls(base_path=base_path, notebook=notebook)
        sidebar = phy_monitor.build_sidebar()
        return pn.Row(sidebar, phy_monitor.build_phy_pane(widget_widths))


def run_dashboard_phy() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    phy_pane = GedMonitoring.display_phy(config.base, args.widget_widths)
    print("Starting Phy. Monitoring on port ", args.port)  # noqa: T201
    pn.serve(phy_pane, port=args.port, show=False)



from __future__ import annotations

import argparse
from pathlib import Path

import importlib.resources

import panel as pn

from legenddashboard.base import Monitoring
from legenddashboard.geds.cal.cal_monitoring import CalMonitoring
from legenddashboard.geds.ged_monitoring import GedMonitoring
from legenddashboard.geds.phy.phy_monitoring import PhyMonitoring
from legenddashboard.llama.llama_monitoring import LlamaMonitoring
from legenddashboard.muon.muon_monitoring import MuonMonitoring
from legenddashboard.spms.sipm_monitoring import SiPMMonitoring
from legenddashboard.util import read_config


def build_dashboard(
    config: str | dict,
    widget_widths: int = 140,
    disable_page: list[str] | None = None,
):
    config = read_config(config)

    # path to period data
    data_path = config.base
    # path to calibration data
    cal_path = config.cal
    # path to physics data
    phy_path = config.phy
    # path to sipm data
    sipm_path = config.sipm
    # path to muon data
    muon_path = config.muon
    # tmp path for caching
    tmp_cal_path = config.tmp
    # llama data path
    llama_path = config.llama

    # use Golden layout template with header and LEGEND logo from LEGEND webpage
    l200_monitoring = pn.template.GoldenTemplate(
        header_background="#f8f8fa",
        header_color="#1A2A5B",
        title="L200 Monitoring Dashboard",
        sidebar_width=7,
        site="",
        logo="https://legend-exp.org/typo3conf/ext/sitepackage/Resources/Public/Images/Logo/logo_legend_tag_next.svg",
        favicon="https://legend-exp.org/typo3conf/ext/sitepackage/Resources/Public/Favicons/android-chrome-96x96.png",
        site_url="https://legend.edm.nat.tum.de/l200_monitoring_auto/",
    )
    # needed to set header title color
    custom_header_title_css = """
    #header {padding: 0}
    .title {
        font-weight: bold;
        font-family: bradley hand;, cursive
        padding-left: 10px;
        color: #1A2A5B;
    }
    """
    l200_monitoring.config.raw_css.append(custom_header_title_css)

    base_monitor = Monitoring(
        base_path=data_path,
        name="L200 Monitoring",
    )
    ged_monitor = GedMonitoring(
        base_path=cal_path,
        run_dict=base_monitor.param.run_dict,
        periods=base_monitor.param.periods,
        period=base_monitor.param.period,
        run=base_monitor.param.run,
        date_range=base_monitor.param.date_range,
        name="L200 Ged Monitoring",
    )
    sidebar = base_monitor.build_sidebar()
    l200_monitoring.sidebar.append(ged_monitor.build_sidebar(sidebar_instance=sidebar))

    if "cal" not in disable_page:
        cal_monitor = CalMonitoring(
            base_path=cal_path,
            tmp_path=tmp_cal_path,
            run_dict=base_monitor.param.run_dict,
            periods=base_monitor.param.periods,
            period=base_monitor.param.period,
            run=base_monitor.param.run,
            date_range=base_monitor.param.date_range,
            channel=ged_monitor.param.channel,
            string=ged_monitor.param.string,
            sort_by=ged_monitor.param.sort_by,
            name="L200 Cal Monitoring",
        )
        cal_panes = cal_monitor.build_cal_panes(
            widget_widths=widget_widths,
        )
        # cal
        for pane in cal_panes.values():
            l200_monitoring.main.append(pane)
    if "phy" not in disable_page:
        if "cal" not in disable_page:
            phy_monitor = PhyMonitoring(
                base_path=cal_path,
                phy_path=phy_path,
                run_dict=base_monitor.param.run_dict,
                periods=base_monitor.param.periods,
                period=base_monitor.param.period,
                run=base_monitor.param.run,
                date_range=base_monitor.param.date_range,
                channel=ged_monitor.param.channel,
                string=ged_monitor.param.string,
                sort_by=ged_monitor.param.sort_by,
                name="L200 Phy Monitoring",
            )
        else:
            phy_monitor = PhyMonitoring(
                phy_path=phy_path,
                base_path=cal_path,
                run_dict=base_monitor.param.run_dict,
                periods=base_monitor.param.periods,
                period=base_monitor.param.period,
                run=base_monitor.param.run,
                date_range=base_monitor.param.date_range,
                name="L200 Phy Monitoring",
            )
        l200_monitoring.main.append(
            phy_monitor.build_phy_pane(
                widget_widths=widget_widths,
            )
        )
    if "spm" not in disable_page:
        sipm_monitor = SiPMMonitoring(
            sipm_path=sipm_path,
            base_path=cal_path,
            run_dict=base_monitor.param.run_dict,
            periods=base_monitor.param.periods,
            period=base_monitor.param.period,
            run=base_monitor.param.run,
            date_range=base_monitor.param.date_range,
            name="L200 SiPM Monitoring",
        )
        l200_monitoring.main.append(
            sipm_monitor.build_spm_pane(
                widget_widths=widget_widths,
            )
        )

    if "muon" not in disable_page:
        muon_monitor = MuonMonitoring(
            muon_path=muon_path,
            base_path=cal_path,
            run_dict=base_monitor.param.run_dict,
            periods=base_monitor.param.periods,
            period=base_monitor.param.period,
            run=base_monitor.param.run,
            date_range=base_monitor.param.date_range,
            name="L200 Muon Monitoring",
        )
        muon_panes = muon_monitor.build_muon_panes(
            widget_widths=widget_widths,
        )
        for pane in muon_panes.values():
            l200_monitoring.main.append(pane)
    if "meta" not in disable_page:
        l200_monitoring.main.append(
            ged_monitor.build_meta_pane(widget_widths=widget_widths)
        )
    if "llama" not in disable_page:
        llama_monitor = LlamaMonitoring(
            llama_path=llama_path,
            base_path=cal_path,
            name="L200 Llama Monitoring",
        )
        l200_monitoring.main.append(
            llama_monitor.build_llama_pane(widget_widths=widget_widths)
        )

    return l200_monitoring


def build_header_logos():
    # Header
    return pn.Row(
        pn.pane.Image(
            "https://legend.edm.nat.tum.de/logos/github-mark.png",
            link_url="https://github.com/legend-exp/",
            fixed_aspect=True,
            width=24,
        ),
        pn.pane.Image(
            "https://legend.edm.nat.tum.de/logos/logo_indico.png",
            link_url="https://indico.legend-exp.org",
            fixed_aspect=True,
            width=24,
        ),
        pn.pane.Image(
            "https://legend.edm.nat.tum.de/logos/confluence.png",
            link_url="https://legend-exp.atlassian.net/wiki/spaces/LEGEND/overview",
            fixed_aspect=True,
            width=24,
        ),
        pn.pane.Image(
            "https://legend.edm.nat.tum.de/logos/elog.png",
            link_url="https://elog.legend-exp.org/ELOG/",
            fixed_aspect=True,
            width=30,
        ),
        width=1800,
        align="center",
    )


def build_info_pane(info_path):
    with Path(info_path).open() as f:
        general_information = f.read()
    return pn.pane.Markdown(general_information)


def run_dashboard() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    argparser.add_argument(
        "-d", "--disable_page", nargs="*", required=False, default=[]
    )
    args = argparser.parse_args()
    
    info_path = importlib.resources.files("legenddashboard") / "information" / "general.md"

    l200_monitoring = build_dashboard(
        args.config_file, args.widget_widths, args.disable_page
    )

    l200_monitoring.header.append(
        pn.Row(pn.Spacer(width=120), build_header_logos(), sizing_mode="stretch_width")
    )

    l200_monitoring.main.append(pn.Tabs(("Information", build_info_pane(info_path))))
    print("Starting Monitoring Dashboard on port ", args.port)  # noqa: T201
    pn.serve(l200_monitoring, port=args.port, show=False)



from __future__ import annotations

import math

import bokeh.palettes as pal
import numpy as np
from bokeh.models import (
    BasicTicker,
    ColorBar,
    ColumnDataSource,
    CustomJSTickFormatter,
    FixedTicker,
    LabelSet,
    LinearColorMapper,
)
from bokeh.plotting import figure

# # display bokeh plots in a notebook
# from bokeh.io import output_notebook
# from bokeh.resources import INLINE
# output_notebook(INLINE)

# functions needed to plot the detectors in their correct shape


def is_coax(d):
    return d["type"] == "coax"


def is_taper(f):
    return f not in (
        {"angle_in_deg": 0, "height_in_mm": 0},
        {
            "radius_in_mm": 0,
            "height_in_mm": 0,
        },
    )


def is_bulletized(f):
    return "bulletization" in f and f["bulletization"] != {
        "top_radius_in_mm": 0,
        "bottom_radius_in_mm": 0,
        "borehole_radius_in_mm": 0,
        "contact_radius_in_mm": 0,
    }


def has_groove(f):
    return "groove" in f and f["groove"] != {
        "outer_radius_in_mm": 0,
        "depth_in_mm": 0,
        "width_in_mm": 0,
    }


def has_borehole(f):
    return "borehole" in f and f["borehole"] != {"gap_in_mm": 0, "radius_in_mm": 0}


def plot_geometry(d, R, H):
    coax = is_coax(d)

    g = d["geometry"]

    DH = g["height_in_mm"]
    DR = g["radius_in_mm"]

    xbot = []
    ybot = []

    # botout = g['taper']['bottom']['outer']
    botout = g["taper"]["bottom"]
    if is_taper(botout):
        TH = botout["height_in_mm"]
        TR = (
            botout["radius_in_mm"]
            if "radius_in_mm" in botout
            else TH * math.sin(botout["angle_in_deg"] * math.pi / 180)
        )
        xbot.extend([DR, DR - TR])
        ybot.extend([H - DH + TH, H - DH])
    else:
        xbot.append(DR)
        ybot.append(H - DH)

    if has_groove(g):
        # GR = g['groove']['outer_radius_in_mm']
        GR = g["groove"]["radius_in_mm"]["outer"]
        GH = g["groove"]["depth_in_mm"]
        # GW = g['groove']['width_in_mm']
        GW = g["groove"]["radius_in_mm"]["outer"] - g["groove"]["radius_in_mm"]["inner"]
        xbot.extend([GR, GR, GR - GW, GR - GW])
        ybot.extend([H - DH, H - DH + GH, H - DH + GH, H - DH])

    if coax:
        BG = g["borehole"]["depth_in_mm"]
        BR = g["borehole"]["radius_in_mm"]
        xbot.extend([BR, BR])
        ybot.extend([H - DH, H - DH + BG])

    xtop = []
    ytop = []

    # topout = g['taper']['top']['outer']
    topout = g["taper"]["top"]
    if is_taper(topout):
        TH = topout["height_in_mm"]
        TR = TH * math.sin(topout["angle_in_deg"] * math.pi / 180)
        xtop.extend([DR, DR - TR])
        ytop.extend([H - TH, H])
    else:
        xtop.append(DR)
        ytop.append(H)

    if has_borehole(g) and not coax:
        BG = g["borehole"]["depth_in_mm"]
        BR = g["borehole"]["radius_in_mm"]

        # topin  = g['taper']['top']['inner']
        topin = g["taper"]["top"]
        if is_taper(topin):
            TH = topin["height_in_mm"]
            TR = TH * math.sin(topin["angle_in_deg"] * math.pi / 180)
            xtop.extend([BR + TR, BR, BR])
            ytop.extend([H, H - TH, H - DH + BG])
        else:
            xtop.extend([BR, BR])
            ytop.extend([H, H - DH + BG])

    x = np.hstack(
        (
            [-x + R for x in xbot],
            [x + R for x in xbot[::-1]],
            [x + R for x in xtop],
            [-x + R for x in xtop[::-1]],
        )
    )
    y = np.hstack((ybot, ybot[::-1], ytop, ytop[::-1]))
    return x, y


def get_plot_source_and_xlabels(
    chan_dict, channel_map, strings_dict, delta_r=160, delta_h=40
):
    xs = []
    ys = []
    ch = []
    hw = []
    sw = []
    dn = []
    st = []
    pos = []
    ax = []
    ay = []

    maxH = 0
    R = 0
    H = 0

    xlabels = {}

    for name, string in strings_dict.items():
        xlabels[R] = name

        for channel_no in string:
            ch.append(channel_no)
            status = chan_dict[channel_map[channel_no]["name"]]
            hw.append(status["usability"])
            sw.append(status["processable"])
            channel = channel_map[channel_no]
            dn.append(channel["name"])
            st.append(channel["location"]["string"])
            pos.append(channel["location"]["position"])
            x, y = plot_geometry(channel, R, H)
            xs.append(x)
            ys.append(y)
            H -= channel["geometry"]["height_in_mm"] + delta_h
            ax.append(R)
            ay.append(H + 10)
        R += delta_r
        maxH = min(H, maxH)
        H = 0

    return ColumnDataSource(
        {
            "xs": xs,  # x-coordinates (needed for plotting)
            "ys": ys,  # y-coordinates (needed for plotting)
            "ch": ch,  # channel numbers
            "hw": hw,  # hardware status
            "sw": sw,  # software status
            "dn": dn,  # detector name
            "st": st,  # string number
            "pos": pos,  # position in string
            "ax": ax,
            "ay": ay,
            "mass": [
                f"{float(channel_map[i]['production']['mass_in_g'])}g" for i in ch
            ],
        }
    ), xlabels


def create_detector_plot(
    source,
    display_dict,
    xlabels,
    ctitle="",
    plot_title="LEGEND detector monitoring",
    palette=None,
    ticker=None,
    formatter=None,
    colour_max=None,
    colour_min=None,
    boolean_scale=False,
):
    if palette is None:
        palette = pal.inferno(256)

    source.data["y_label"] = [display_dict[ch] for ch in source.data["dn"]]
    tooltips = [
        ("Detector Name", "@dn"),
        ("Channel", "@ch"),
        ("String", "@st"),
        ("Usability", "@hw"),
        ("Processable", "@sw"),
        ("Mass", "@mass"),
        (f"{ctitle}", "@y_label"),
    ]

    if ticker is None:
        ticker = BasicTicker()

    # To Do: find optimal width and height values for plotting (do not hardcode)
    p = figure(
        title=plot_title,
        width=1200,
        height=920,
        tools="pan,box_zoom,hover,reset,save",
        tooltips=tooltips,
        match_aspect=True,
    )

    # handle colors according to display_dict
    values = [val for val in display_dict.values() if val is not None]
    minvalue = min(values) if colour_min is None else colour_min
    maxvalue = max(values) if colour_min is None else colour_max

    color_mapper = LinearColorMapper(palette=palette, low=minvalue, high=maxvalue)

    def convert_value_to_colour(v):
        if boolean_scale:
            return palette[int(v)]
        if v is None or math.isnan(v) or v == 0:
            return "white"
        if v > maxvalue:
            return palette[-1]
        if v < maxvalue:
            return palette[0]
        return palette[int((v - minvalue) / (maxvalue - minvalue) * (len(palette) - 1))]

    colors = list(map(convert_value_to_colour, display_dict.values()))
    source.data["color"] = colors

    # plot detector geometries with respective colors
    p.patches("xs", "ys", source=source, line_color="black", color="color")
    color_bar = ColorBar(color_mapper=color_mapper, ticker=ticker, title=ctitle)
    if formatter is not None:
        color_bar.formatter = formatter
    p.add_layout(color_bar, "right")

    # annotate detector names underneath the detectors
    labels = LabelSet(
        x="ax",
        y="ay",
        text="dn",
        source=source,
        text_font_size="10px",
        text_align="center",
    )
    p.add_layout(labels)

    # remove unnecessary lines in the plot
    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    p.xaxis.ticker = list(xlabels)
    p.xaxis.major_label_overrides = xlabels
    p.yaxis.visible = False
    p.xaxis.major_label_text_font_style = "bold"
    p.xaxis.axis_label_text_font_size = "25px"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    # return the plot
    return p


def plot_visu_usability(source, chan_dict, channel_map, xlabels):
    color_dict = {"on": 2, "off": 0, "ac": 1}
    display_dict = {i: color_dict[chan_dict[i]["usability"]] for i in source.data["dn"]}
    palette = ("red", "orange", "green")
    ctitle = "Usability"
    ticker = FixedTicker(ticks=[0.3, 1.0, 1.7], tags=["red", "orange", "green"])
    formatter = CustomJSTickFormatter(
        code="""
        var mapping = {0.3: "off", 1.0: "ac", 1.7: "on"};
        return mapping[tick];
    """
    )
    return create_detector_plot(
        source,
        display_dict,
        xlabels,
        ctitle=ctitle,
        palette=palette,
        ticker=ticker,
        formatter=formatter,
        boolean_scale=True,
    )


def plot_visu_processable(source, chan_dict, channel_map, xlabels):
    color_dict = {True: 1, False: 0}
    display_dict = {
        i: color_dict[chan_dict[i]["processable"]] for i in source.data["dn"]
    }
    palette = ("red", "green")
    ctitle = "Processable"
    ticker = FixedTicker(ticks=[0.25, 0.75], tags=["True", "False"])
    formatter = CustomJSTickFormatter(
        code="""
        var mapping = {0.25: "True", 0.75: "False"};
        return mapping[tick];
    """
    )
    return create_detector_plot(
        source,
        display_dict,
        xlabels,
        ctitle=ctitle,
        palette=palette,
        ticker=ticker,
        formatter=formatter,
        boolean_scale=True,
    )


def plot_visu_mass(source, chan_dict, channel_map, xlabels):
    display_dict = {
        i: channel_map[i]["production"]["mass_in_g"] for i in source.data["dn"]
    }
    ctitle = "Mass in g"
    return create_detector_plot(source, display_dict, xlabels, ctitle=ctitle)


def plot_visu_depletion(source, chan_dict, channel_map, xlabels):
    display_dict = {
        i: channel_map[i]["characterization"]["manufacturer"]["depletion_voltage_in_V"]
        for i in source.data["dn"]
    }
    ctitle = "Depletion voltage in V (manufacturer)"
    palette = pal.viridis(256)
    return create_detector_plot(
        source, display_dict, xlabels, ctitle=ctitle, palette=palette
    )


def plot_visu_operation(source, chan_dict, channel_map, xlabels):
    display_dict = {
        i: channel_map[i]["characterization"]["manufacturer"][
            "recommended_voltage_in_V"
        ]
        for i in source.data["dn"]
    }
    ctitle = "Operational voltage in V (manufacturer)"
    palette = pal.viridis(256)
    return create_detector_plot(
        source, display_dict, xlabels, ctitle=ctitle, palette=palette
    )


def plot_visu_enrichment(source, chan_dict, channel_map, xlabels):
    def get_enrichment(channel):
        try:
            ret = channel_map[channel]["production"]["enrichment"]["val"]
        except KeyError:
            ret = 0.0
        return ret

    display_dict = {i: get_enrichment(i) for i in source.data["dn"]}
    ctitle = "Enrichment in %"
    return create_detector_plot(source, display_dict, xlabels, ctitle=ctitle)



from __future__ import annotations

import bisect
import datetime as dtt
import logging
import time
from datetime import date, datetime
from pathlib import Path

import numpy as np
import panel as pn
import param
from bokeh.io import output_notebook
from bokeh.resources import INLINE
from dbetto import Props

from legenddashboard.util import gen_run_dict, logo_path, sort_dets

log = logging.getLogger(__name__)


class Monitoring(param.Parameterized):
    """
    Base class for monitoring dashboards.
    """

    base_path = param.String("", allow_refs=True, nested_refs=True)
    prod_config = param.Dict({}, allow_refs=True, nested_refs=True)
    tier_dict = param.Dict({}, allow_refs=True, nested_refs=True)
    period = param.Selector(
        default="p00",
        objects=[f"p{i:02}" for i in range(100)],
        allow_refs=True,
        nested_refs=True,
    )
    run = param.Selector(
        default="r000",
        objects=[f"r{i:03}" for i in range(100)],
        allow_refs=True,
        nested_refs=True,
    )
    run_dict = param.Dict({}, allow_refs=True, nested_refs=True)
    periods = param.Dict({}, allow_refs=True, nested_refs=True)

    date_range = param.DateRange(
        default=(
            datetime.now() - dtt.timedelta(minutes=10),
            datetime.now() + dtt.timedelta(minutes=10),
        ),
        bounds=(
            datetime(2000, 1, 1, 0, 0, 0),
            datetime(2100, 1, 1, 0, 0, 0),
        ),
        allow_refs=True,
        nested_refs=True,
    )

    def __init__(self, base_path, notebook=False, **params):
        if notebook is True:
            output_notebook(INLINE)
        self.cached_plots = {}
        self.base_path = base_path
        self.sort_obj = sort_dets(base_path)

        super().__init__(**params)

        self.tier_dict = {
            "raw": "raw",
            "tcm": "tcm",
            "dsp": "dsp",
            "hit": "hit",
            "evt": "evt",
        }

        if "ref-v" in str(self.base_path):
            self.tier_dict["dsp"] = "psp"
            self.tier_dict["hit"] = "pht"
            self.tier_dict["evt"] = "pet"

        prod_config = Path(self.base_path) / "dataflow-config.yaml"
        self.prod_config = Props.read_from(prod_config, subst_pathvar=True)
        if self.period == "p00":
            self.periods = gen_run_dict(self.base_path)
            log.debug("updating")
            self.param["period"].objects = list(self.periods)
            self.period = list(self.periods)[-1]
            self._get_period_data(None)

        self.param.watch(self._get_period_data, ["period"], precedence=0)
        self.param.watch(self._get_run_dict, ["date_range"], precedence=0)

    def _get_period_data(self, event=None):  # noqa: ARG002
        self.run_dict = self.periods[self.period]

        self.param["run"].objects = list(self.run_dict)
        if self.run == list(self.run_dict)[-1]:
            self.run = next(iter(self.run_dict))
        else:
            self.run = list(self.run_dict)[-1]

        start_period = sorted(self.periods)[0]
        start_run = sorted(self.periods[start_period])[0]
        end_period = sorted(self.periods)[-1]
        end_run = sorted(self.periods[end_period])[-1]

        self.param["date_range"].bounds = (
            datetime.strptime(
                self.periods[start_period][start_run]["timestamp"], "%Y%m%dT%H%M%SZ"
            )
            - dtt.timedelta(minutes=100),
            datetime.strptime(
                self.periods[end_period][end_run]["timestamp"], "%Y%m%dT%H%M%SZ"
            )
            + dtt.timedelta(minutes=110),
        )
        self.date_range = (
            datetime.strptime(
                self.periods[start_period][start_run]["timestamp"], "%Y%m%dT%H%M%SZ"
            )
            - dtt.timedelta(minutes=100),
            datetime.strptime(
                self.periods[end_period][end_run]["timestamp"], "%Y%m%dT%H%M%SZ"
            )
            + dtt.timedelta(minutes=110),
        )

    def _get_run_dict(self, event=None):  # noqa: ARG002
        start_time = time.time()
        valid_from = [
            datetime.timestamp(
                datetime.strptime(self.run_dict[entry]["timestamp"], "%Y%m%dT%H%M%SZ")
            )
            for entry in self.run_dict
        ]
        if isinstance(self.date_range[0], date):
            low_range = datetime.timestamp(
                datetime.combine(self.date_range[0], datetime.min.time())
            )
        else:
            low_range = datetime.timestamp(self.date_range[0])
        if isinstance(self.date_range[0], date):
            high_range = datetime.timestamp(
                datetime.combine(self.date_range[1], datetime.max.time())
            )
        else:
            high_range = datetime.timestamp(self.date_range[1])
        pos1 = bisect.bisect_right(valid_from, low_range)
        pos2 = bisect.bisect_left(valid_from, high_range)
        pos1 = max(pos1, 0)
        pos2 = min(len(self.run_dict), pos2)
        valid_idxs = np.arange(pos1, pos2, 1)
        valid_keys = np.array(list(self.run_dict))[valid_idxs]
        out_dict = {key: self.run_dict[key] for key in valid_keys}
        log.debug("Time to get run dict:", extra={"time": time.time() - start_time})
        return out_dict

    def build_sidebar(self):
        run_param = pn.widgets.MenuButton(
            name=f"Run {int(self.run[1:]):02d}",
            button_type="primary",
            sizing_mode="stretch_width",
            items=self.param.run.objects,
        )

        def update_run(event):
            self.run = event.new
            run_param.name = f"Run {int(self.run[1:]):02d}"

        run_param.on_click(update_run)
        # run_param        = pn.Param(self.param, widgets={'run': {'widget_type': pn.widgets.Select, 'width': 100}}, parameters=['run'], show_labels=False, show_name=False, design=Bootstrap)
        period_param = pn.widgets.MenuButton(
            name=f"Period {int(self.period[1:]):02d}",
            button_type="primary",
            sizing_mode="stretch_width",
            items=self.param.period.objects,
        )

        def update_period(event):
            self.period = event.new
            run_param.items = self.param.run.objects
            run_param.name = f"Run {int(self.run[1:]):02d}"
            period_param.name = f"Period {int(self.period[1:]):02d}"

        period_param.on_click(update_period)

        return pn.Column(
            pn.pane.SVG(
                logo_path / "Period.svg",
                height=25,
            ),
            period_param,
            pn.pane.SVG(
                logo_path / "Run.svg",
                height=25,
            ),
            run_param,
            sizing_mode="stretch_width",
        )

    def build_base(path, notebook=False):
        monitor = Monitoring(
            base_path=path,
            notebook=notebook,
        )

        return monitor.build_sidebar()



from __future__ import annotations

import argparse
import logging
import pickle as pkl
import shelve
import time
from pathlib import Path

import matplotlib.pyplot as plt
import panel as pn
import param

import legenddashboard.geds.string_visulization as visu
from legenddashboard.geds import cal
from legenddashboard.geds.ged_monitoring import GedMonitoring
from legenddashboard.util import logo_path, read_config, sorter

log = logging.getLogger(__name__)

# calibration plots
plt.rcParams["font.size"] = 10
plt.rcParams["figure.figsize"] = (16, 6)
plt.rcParams["figure.dpi"] = 100


class CalMonitoring(GedMonitoring):
    cached_data = param.Dict(default={"hit": {}, "dsp": {}})
    tmp_path = param.String("/tmp/")
    plot_type_tracking = param.ObjectSelector(
        default=list(cal.tracking_plots)[1],
        objects=list(cal.tracking_plots),
    )

    parameter = param.ObjectSelector(
        default=next(iter(cal.all_detailed_plots)), objects=list(cal.all_detailed_plots)
    )

    plot_type_details = param.ObjectSelector(
        default=cal.detailed_plots[0], objects=cal.detailed_plots
    )

    plot_type_summary = param.ObjectSelector(
        default=list(cal.summary_plots)[3],
        objects=list(cal.summary_plots),
    )
    plot_types_download = param.Selector(
        objects=["FWHM Qbb", "FWHM FEP", "A/E", "PZ", "Alpha"],
        default="FWHM Qbb",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.update_plot_dict(None)
        self.param.watch(
            self.download_summary_files,
            ["period", "run", "sort_by", "plot_types_download"],
            precedence=3,
            queued=True,
        )
        self.param.watch(
            self.view_summary,
            ["period", "run", "sort_by", "plot_type_summary"],
            precedence=2,
            queued=True,
        )
        self.param.watch(
            self.update_plot_dict, ["period", "run"], precedence=2, queued=True
        )
        self.param.watch(
            self.update_channel_plot_dict, ["channel"], precedence=2, queued=True
        )
        self.param.watch(
            self._clear_cached_data,
            [
                "period",
            ],
            precedence=1,
            queued=True,
        )
        self.param.watch(
            self.view_tracking,
            ["period", "date_range", "sort_by", "string", "plot_type_tracking"],
            precedence=2,
            queued=True,
        )
        self.param.watch(
            self.view_details,
            ["period", "run", "channel", "parameter", "plot_type_details"],
            precedence=2,
            queued=True,
        )

    def download_summary_files(self, event=None):  # noqa: ARG002
        start_time = time.time()
        try:
            download_file, download_filename = cal.summary_plots[
                self.plot_types_download
            ](
                self.prod_config,
                self.run,
                self.run_dict[self.run],
                self.base_path,
                self.period,
                key=self.sort_by,
                download=True,
                sort_dets_obj=self.sort_obj,
                cache_data=self.cached_data,
            )
            # log.debug(download_filename)
            tmp_path = Path(self.tmp_path)
            if not (tmp_path / download_filename).exists():
                download_file.to_csv(tmp_path / download_filename, index=False)
                log.debug(download_file, tmp_path)
            ret = pn.widgets.FileDownload(
                tmp_path / download_filename,
                filename=download_filename,
                button_type="success",
                embed=False,
                name="Click to download 'csv'",
                width=350,
            )
        except BaseException:
            ret = pn.widgets.FileDownload(
                None,
                filename="temp",
                button_type="success",
                embed=False,
                name="Click to download 'csv'",
                width=350,
            )
        log.debug(
            "Time to download summary files:", extra={"time": time.time() - start_time}
        )
        return ret

    def view_summary(self, event=None):  # noqa: ARG002
        start_time = time.time()
        figure = None
        try:
            if self.cached_data == {}:
                self.cached_data = {
                    "hit": {},
                    "dsp": {},
                }
            if self.plot_type_summary in [
                "FWHM Qbb",
                "FWHM FEP",
                "Energy Residuals",
                "A/E Status",
                "PZ",
                "CT Alpha",
                "Valid. E",
                "Valid. A/E",
                "A/E SF",
            ]:
                figure = cal.summary_plots[self.plot_type_summary](
                    self.prod_config,
                    self.run,
                    self.run_dict[self.run],
                    self.base_path,
                    self.period,
                    key=self.sort_by,
                    sort_dets_obj=self.sort_obj,
                    cache_data=self.cached_data,
                )

            elif self.plot_type_summary in ["Detector Status", "FEP Counts"]:
                # elif self.plot_type_summary in ["Detector Status"]:
                strings_dict, meta_visu_chan_dict, meta_visu_channel_map = sorter(
                    self.base_path,
                    self.run_dict[self.run]["timestamp"],
                    key="String",
                    sort_dets_obj=self.sort_obj,
                )
                meta_visu_source, meta_visu_xlabels = visu.get_plot_source_and_xlabels(
                    meta_visu_chan_dict, meta_visu_channel_map, strings_dict
                )
                # self.meta_visu_chan_dict, self.meta_visu_channel_map = chan_dict, channel_map
                figure = cal.summary_plots[self.plot_type_summary](
                    self.prod_config,
                    self.run,
                    self.run_dict[self.run],
                    self.base_path,
                    meta_visu_source,
                    meta_visu_xlabels,
                    self.period,
                    key=self.sort_by,
                    sort_dets_obj=self.sort_obj,
                    cache_data=self.cached_data,
                )
            elif self.plot_type_summary in [
                "Baseline Spectrum",
                "Energy Spectrum",
                "Baseline Stability",
                "FEP Stability",
                "Pulser Stability",
            ]:
                figure = cal.summary_plots[self.plot_type_summary](
                    self.prod_config,
                    self.common_dict,
                    self.channel_map,
                    self.strings_dict[self.string],
                    self.string,
                    self.run,
                    self.period,
                    self.run_dict[self.run],
                    key=self.sort_by,
                    sort_dets_obj=self.sort_obj,
                    cache_data=self.cached_data,
                )
            else:
                figure = plt.figure()
        except BaseException:
            pass
        log.debug("Time to get summary plot:", extra={"time": time.time() - start_time})
        return figure

    def _clear_cached_data(self, event=None):  # noqa: ARG002
        """
        Clear the cached data if the period changes.
        """
        self.cached_data["hit"] = {}
        self.cached_data["dsp"] = {}

    def view_tracking(self, event=None):  # noqa: ARG002
        figure = None
        try:
            if self.plot_type_tracking != "Energy Residuals":
                figure = cal.plot_tracking(
                    self._get_run_dict(),
                    self.base_path,
                    cal.tracking_plots[self.plot_type_tracking],
                    self.string,
                    self.period,
                    self.plot_type_tracking,
                    key=self.sort_by,
                    cache_data=self.cached_data,
                    sort_dets_obj=self.sort_obj,
                )
            else:
                figure = cal.plot_energy_residuals_period(
                    self._get_run_dict(),
                    self.base_path,
                    self.period,
                    key=self.sort_by,
                    cache_data=self.cached_data,
                    sort_dets_obj=self.sort_obj,
                )
        except BaseException:
            pass
        return figure

    def update_plot_dict(self, event=None):  # noqa: ARG002
        start_time = time.time()
        self.plot_dict = (
            Path(self.prod_config["paths"]["plt"])
            / f"hit/cal/{self.period}/{self.run}"
            / f'{self.run_dict[self.run]["experiment"]}-{self.period}-{self.run}-cal-{self.run_dict[self.run]["timestamp"]}-plt_hit'
        )

        # log.debug(self.run_dict)
        # log.debug(self.plot_dict)
        with shelve.open(self.plot_dict, "r", protocol=pkl.HIGHEST_PROTOCOL) as shelf:
            channels = list(shelf.keys())

        with shelve.open(self.plot_dict, "r", protocol=pkl.HIGHEST_PROTOCOL) as shelf:
            self.common_dict = shelf["common"]
        channels.remove("common")
        self.strings_dict, self.chan_dict, self.channel_map = sorter(
            self.base_path,
            self.run_dict[self.run]["timestamp"],
            "String",
            sort_dets_obj=self.sort_obj,
        )

        self.param["channel"].objects = channels
        self.channel = channels[0]

        self.update_strings()
        self.update_channel_plot_dict()
        log.debug("Time to update plot dict:", extra={"time": time.time() - start_time})

    def update_channel_plot_dict(self, event=None):  # noqa: ARG002
        start_time = time.time()
        log.debug(self.channel)
        with shelve.open(self.plot_dict, "r", protocol=pkl.HIGHEST_PROTOCOL) as shelf:
            self.plot_dict_ch = shelf[self.channel[:9]]
        with shelve.open(
            str(self.plot_dict).replace("hit", "dsp"),
            "r",
            protocol=pkl.HIGHEST_PROTOCOL,
        ) as shelf:
            self.dsp_dict = shelf[self.channel[:9]]
        log.debug(
            "Time to update channel plot dict:",
            extra={"time": time.time() - start_time},
        )

    @param.depends("parameter", watch=True)
    def update_plot_type_details(self):
        start_time = time.time()
        plots = cal.all_detailed_plots[self.parameter]
        self.param["plot_type_details"].objects = plots
        self.plot_type_details = plots[0]
        log.debug(
            "Time to update plot type details:",
            extra={"time": time.time() - start_time},
        )

    def view_details(self, event=None):  # noqa: ARG002
        fig_pane = pn.pane.Matplotlib(plt.figure(), sizing_mode="scale_width")
        try:
            if self.parameter == "A/E":
                fig = self.plot_dict_ch["aoe"][self.plot_type_details]
                dummy = plt.figure()
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.parameter == "Baseline":
                fig = self.plot_dict_ch["ecal"][self.plot_type_details]
                dummy = plt.figure()
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.parameter == "PZ":
                fig = self.dsp_dict["pz"][self.plot_type_details]
                dummy = plt.figure()
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.parameter == "Optimisation":
                fig = self.dsp_dict[
                    f"{self.plot_type_details.split('_')[0]}_optimisation"
                ][f"{self.plot_type_details.split('_')[1]}_space"]
                dummy = plt.figure()
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.plot_type_details in {"spectrum", "logged_spectrum"}:
                fig = cal.plot_spectrum(
                    self.plot_dict_ch["ecal"][self.parameter]["spectrum"],
                    self.channel,
                    log=self.plot_type_details != "spectrum",
                )
                fig_pane = fig
            elif self.plot_type_details == "survival_frac":
                fig = cal.plot_survival_frac(
                    self.plot_dict_ch["ecal"][self.parameter]["survival_frac"]
                )
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.plot_type_details == "cut_spectrum":
                fig = cal.plot_cut_spectra(
                    self.plot_dict_ch["ecal"][self.parameter]["spectrum"]
                )
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            elif self.plot_type_details == "peak_track":
                fig = cal.track_peaks(self.plot_dict_ch["ecal"][self.parameter])
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
            else:
                fig = self.plot_dict_ch["ecal"][self.parameter][self.plot_type_details]
                dummy = plt.figure()
                new_manager = dummy.canvas.manager
                new_manager.canvas.figure = fig
                fig.set_canvas(new_manager.canvas)
                fig_pane = pn.pane.Matplotlib(fig, sizing_mode="scale_width")
        except BaseException:
            pass
        return fig_pane

    def build_detailed_pane(self, widget_widths: int = 140):
        details_ch_param = pn.Param(
            self.param,
            widgets={
                "channel": {"widget_type": pn.widgets.Select, "width": widget_widths}
            },
            parameters=["channel"],
            show_labels=False,
            show_name=False,
            sort=True,
        )

        details_type_param = pn.Param(
            self.param,
            widgets={
                "plot_type_details": {
                    "widget_type": pn.widgets.Select,
                    "width": widget_widths,
                }
            },
            # 'plot_type_details': {'widget_type': pn.widgets.RadioButtonGroup, 'button_type': 'success',
            #         'orientation':"vertical", 'width': widget_widths}},
            parameters=["plot_type_details"],
            show_labels=False,
            show_name=False,
            sort=True,
        )

        details_param_currentValue = pn.pane.Markdown(f"## {self.parameter}")
        details_param = pn.widgets.MenuButton(
            name="Detailed Plots",
            button_type="primary",
            width=widget_widths,
            items=self.param.parameter.objects,
        )

        def update_details_plots(event):
            self.parameter = event.new
            details_param_currentValue.object = f"## {event.new}"

        details_param.on_click(update_details_plots)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Calibration.svg",
                    height=25,
                ),
                details_param,
            ),
            pn.Row("## Current Plot:", details_param_currentValue),
            pn.Row("Channel:", details_ch_param, "Plot type:", details_type_param),
            self.get_run_and_channel,
            self.view_details,
            name="Cal. Details",
            sizing_mode="scale_both",
        )

    def build_summary_pane(self, widget_widths: int = 140):
        summary_param_currentValue = pn.pane.Markdown(f"## {self.plot_type_summary}")
        summary_param = pn.widgets.MenuButton(
            name="Summary Plots",
            button_type="primary",
            width=widget_widths,
            items=self.param.plot_type_summary.objects,
        )

        def update_summary_plots(event):
            self.plot_type_summary = event.new
            summary_param_currentValue.object = f"## {event.new}"

        summary_param.on_click(update_summary_plots)

        summary_param_download = pn.Param(
            self.param,
            widgets={
                "plot_types_download": {
                    "widget_type": pn.widgets.Select,
                    "width": widget_widths,
                }
            },
            parameters=["plot_types_download"],
            show_labels=False,
            show_name=False,
        )
        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Calibration.svg",
                    height=25,
                ),
                summary_param,
            ),
            pn.Row("## Current Plot:", summary_param_currentValue),
            "Download Raw Data",
            pn.Row(summary_param_download, self.download_summary_files),
            self.view_summary,  # , sizing_mode="scale_both"),
            name="Cal. Summary",
            sizing_mode="scale_both",
        )

    def build_tracking_pane(self, widget_widths: int = 140):
        # tracking_range_param = pn.Param(
        #     self.param,
        #     widgets={
        #         "date_range": {
        #             "widget_type": pn.widgets.DatetimeRangePicker,
        #             "width": widget_widths,
        #             "enable_time": False,
        #             "enable_seconds": False,
        #         }
        #     },
        #     parameters=["date_range"],
        #     show_labels=False,
        #     show_name=False,
        # )

        tracking_param_currentValue = pn.pane.Markdown(f"## {self.plot_type_tracking}")
        tracking_param = pn.widgets.MenuButton(
            name="Tracking Plots",
            button_type="primary",
            width=widget_widths,
            items=self.param.plot_type_tracking.objects,
        )

        def update_tracking_plots(event):
            self.plot_type_tracking = event.new
            tracking_param_currentValue.object = f"## {event.new}"

        tracking_param.on_click(update_tracking_plots)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Calibration.svg",
                    height=25,
                ),
                tracking_param,
            ),
            pn.Row("## Current Plot:", tracking_param_currentValue),
            # pn.Row("Selected time range:", tracking_range_param),
            self.view_tracking,
            name="Cal. Tracking",
            sizing_mode="scale_both",
        )

    def build_cal_panes(self, widget_widths: int = 140):
        self.update_plot_dict()
        return {
            "Cal. Summary": self.build_summary_pane(widget_widths),
            "Cal. Details": self.build_detailed_pane(widget_widths),
            "Cal. Tracking": self.build_tracking_pane(widget_widths),
        }

    @classmethod
    def init_cal_panes(
        cls,
        base_path,
        widget_widths: int = 140,
    ):
        """
        Initialize the calibration panes.

        Args:
            widget_widths (int): Width of the widgets.

        Returns:
            dict: Dictionary containing the calibration panes.
        """
        cal_monitor = cls(base_path=base_path)
        return cal_monitor.build_cal_panes(widget_widths)

    @classmethod
    def display_cal_panes(
        cls,
        base_path,
        notebook=False,
        widget_widths: int = 140,
    ):
        """
        View the calibration panes.

        Args:
            widget_widths (int): Width of the widgets.

        Returns:
            pn.Row: Row containing the calibration panes.
        """
        cal_monitor = cls(base_path=base_path, notebook=notebook)
        sidebar = cal_monitor.build_sidebar()
        return pn.Row(
            sidebar, pn.Tabs(*cal_monitor.build_cal_panes(widget_widths).values())
        )


def run_dashboard_cal() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    cal_panes = CalMonitoring.display_cal_panes(config.base, args.widget_widths)
    print("Starting Cal. Monitoring on port ", args.port)  # noqa: T201
    pn.serve(cal_panes, port=args.port, show=False)
